# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/06_models.roberta.conditional_prob_classifiers.ipynb.

# %% ../../../nbs/06_models.roberta.conditional_prob_classifiers.ipynb 3
from __future__ import annotations
import numpy as np
import torch
from transformers.models.roberta.configuration_roberta import RobertaConfig
from transformers.models.roberta.modeling_roberta import RobertaModel
from sklearn.preprocessing import MultiLabelBinarizer
from transformers.modeling_outputs import SequenceClassifierOutput
from transformers.models.roberta.modeling_roberta import RobertaPreTrainedModel

# %% auto 0
__all__ = ['build_standard_condition_mask', 'RobertaHSCCProbSequenceClassification']

# %% ../../../nbs/06_models.roberta.conditional_prob_classifiers.ipynb 5
def build_standard_condition_mask(df_labels,
                                  label1,label2):
    L1_SIZE = df_labels[label1].nunique()
    L2_SIZE = df_labels[label2].nunique()

    
    df_labels = df_labels.drop_duplicates().sort_values([label1,label2])
    _d = df_labels.groupby([label1])[label2].apply(list).to_dict()
    
    mask_l1 = torch.eye(L1_SIZE) ==1
    mlb = MultiLabelBinarizer()
    mlb.fit([np.arange(L2_SIZE)])
    mask_l2 = torch.tensor(mlb.transform(list(_d.values())) == 1)
    
    mask_final= torch.cat((mask_l1,mask_l2),1)
    
    return mask_final

# %% ../../../nbs/06_models.roberta.conditional_prob_classifiers.ipynb 8
class RobertaHSCCProbSequenceClassification(RobertaPreTrainedModel):
    """
    Roberta Conditional Probability Architecture with Hidden-State-Concatenation for Sequence Classification task
    """
    config_class = RobertaConfig

    def __init__(self, 
                 config, # HuggingFace model configuration
                 size_l1=None, # Number of classes for head 1
                 size_l2=None, # Number of classes for head 2
                 standard_mask=None, # Mask for conditional probability
                 layer2concat=4, # number of hidden layer to concatenate (counting from top)
                 device=None, # CPU or GPU
                 head_class=None, # The class object of the head. You can use RobertaClassificationHeadCustom as default
                 **head_class_kwargs, # Keyword arguments for the head class
                ):
        super().__init__(config)
        self.training_device = device if device is not None else torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
        self.size_l1 = size_l1
        self.size_l2 = size_l2
        self.layer2concat=layer2concat
        self.head_class_sizes=[size_l1,size_l2] # will be useful for metric calculation later
        
        # set num_labels for config
        num_labels = size_l1+size_l2
        config.num_labels = num_labels
        
        self.body_model = RobertaModel(config, add_pooling_layer=False)
        self.standard_mask = standard_mask.to(self.training_device)
        self.classification_head = head_class(config=config,layer2concat=layer2concat,
                                              **head_class_kwargs) 


    def forward(self, input_ids=None, attention_mask=None, token_type_ids=None,
                labels=None, 
                output_attentions= None,
                output_hidden_states= None,
                return_dict= None,
                **kwargs):
        # Use model body to get encoder representations
        # the only ones we need for now are input_ids and attention_mask
        output_hidden_states = output_hidden_states if output_hidden_states is not None else self.config.output_hidden_states
        
        outputs = self.body_model(input_ids, 
                                  attention_mask=attention_mask,
                                  token_type_ids=token_type_ids, 
                                  output_attentions=output_attentions,
                                  output_hidden_states=output_hidden_states,
                                  return_dict=return_dict,
                                  **kwargs)
        
        hidden_states = outputs['hidden_states'] # tuples with len 13 (number of layer/block)
        # each with shape: (bs,seq_len,hidden_size_len), e.g. for phobert: (bs,256, 768)
        # Note: hidden_size_len = embedding_size
        
        hidden_concat = torch.cat([hidden_states[i][:,0] for i in range(-1,-self.layer2concat-1,-1)],
                                  -1)
        
        # classification head
        logits = self.classification_head(hidden_concat)

        loss = None
        if labels is not None:
            # labels shape: (bs,2), first is L1, second is L2
            labels_l1 = labels[:,0].view(-1) #(bs,)
            labels_l2 = labels[:,1].view(-1) #(bs,)
            l1_1hot = torch.nn.functional.one_hot(labels_l1, num_classes=self.size_l1)
            l2_1hot = torch.nn.functional.one_hot(labels_l2, num_classes=self.size_l2)
            label_concat_1hot = torch.cat((l1_1hot,l2_1hot),1) # (bs,L1+L2)

            # the original approach: positives and other children of same parents
            _mask = self.standard_mask[labels_l1]
            loss_func = torch.nn.BCEWithLogitsLoss(reduction='none')

            loss = loss_func(logits,label_concat_1hot.float())
            loss = torch.mul(loss,_mask)
            loss = (loss.sum(axis=1)/_mask.sum(axis=1)).mean()
            
        # Return model output object
        return SequenceClassifierOutput(loss=loss, logits=logits,
                                     hidden_states=None,
                                     attentions=outputs.attentions)



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/that-nlp-library',
                'doc_host': 'https://anhquan0412.github.io',
                'git_url': 'https://github.com/anhquan0412/that-nlp-library',
                'lib_path': 'that_nlp_library'},
  'syms': { 'that_nlp_library.evaluations': { 'that_nlp_library.evaluations._show_metrics': ( 'evaluations.html#_show_metrics',
                                                                                              'that_nlp_library/evaluations.py'),
                                              'that_nlp_library.evaluations.evaluate_classification_model': ( 'evaluations.html#evaluate_classification_model',
                                                                                                              'that_nlp_library/evaluations.py'),
                                              'that_nlp_library.evaluations.evaluate_classification_model_metadata': ( 'evaluations.html#evaluate_classification_model_metadata',
                                                                                                                       'that_nlp_library/evaluations.py'),
                                              'that_nlp_library.evaluations.show_top_n_predictions': ( 'evaluations.html#show_top_n_predictions',
                                                                                                       'that_nlp_library/evaluations.py')},
            'that_nlp_library.model_lm_main': { 'that_nlp_library.model_lm_main.ModelLMController': ( 'model_lm_main.html#modellmcontroller',
                                                                                                      'that_nlp_library/model_lm_main.py'),
                                                'that_nlp_library.model_lm_main.ModelLMController.__init__': ( 'model_lm_main.html#modellmcontroller.__init__',
                                                                                                               'that_nlp_library/model_lm_main.py'),
                                                'that_nlp_library.model_lm_main.ModelLMController.fit': ( 'model_lm_main.html#modellmcontroller.fit',
                                                                                                          'that_nlp_library/model_lm_main.py'),
                                                'that_nlp_library.model_lm_main.ModelLMController.get_hidden_states': ( 'model_lm_main.html#modellmcontroller.get_hidden_states',
                                                                                                                        'that_nlp_library/model_lm_main.py'),
                                                'that_nlp_library.model_lm_main.ModelLMController.get_hidden_states_from_raw_dset': ( 'model_lm_main.html#modellmcontroller.get_hidden_states_from_raw_dset',
                                                                                                                                      'that_nlp_library/model_lm_main.py'),
                                                'that_nlp_library.model_lm_main.ModelLMController.get_hidden_states_from_raw_text': ( 'model_lm_main.html#modellmcontroller.get_hidden_states_from_raw_text',
                                                                                                                                      'that_nlp_library/model_lm_main.py'),
                                                'that_nlp_library.model_lm_main.ModelLMController.predict_ddict': ( 'model_lm_main.html#modellmcontroller.predict_ddict',
                                                                                                                    'that_nlp_library/model_lm_main.py'),
                                                'that_nlp_library.model_lm_main.ModelLMController.predict_raw_text': ( 'model_lm_main.html#modellmcontroller.predict_raw_text',
                                                                                                                       'that_nlp_library/model_lm_main.py'),
                                                'that_nlp_library.model_lm_main.compute_lm_accuracy': ( 'model_lm_main.html#compute_lm_accuracy',
                                                                                                        'that_nlp_library/model_lm_main.py'),
                                                'that_nlp_library.model_lm_main.extract_hidden_states': ( 'model_lm_main.html#extract_hidden_states',
                                                                                                          'that_nlp_library/model_lm_main.py'),
                                                'that_nlp_library.model_lm_main.finetune_lm': ( 'model_lm_main.html#finetune_lm',
                                                                                                'that_nlp_library/model_lm_main.py'),
                                                'that_nlp_library.model_lm_main.language_model_init': ( 'model_lm_main.html#language_model_init',
                                                                                                        'that_nlp_library/model_lm_main.py'),
                                                'that_nlp_library.model_lm_main.preprocess_lm_logits_for_metrics': ( 'model_lm_main.html#preprocess_lm_logits_for_metrics',
                                                                                                                     'that_nlp_library/model_lm_main.py')},
            'that_nlp_library.model_main': { 'that_nlp_library.model_main.ModelController': ( 'model_main.html#modelcontroller',
                                                                                              'that_nlp_library/model_main.py'),
                                             'that_nlp_library.model_main.ModelController.__init__': ( 'model_main.html#modelcontroller.__init__',
                                                                                                       'that_nlp_library/model_main.py'),
                                             'that_nlp_library.model_main.ModelController.fit': ( 'model_main.html#modelcontroller.fit',
                                                                                                  'that_nlp_library/model_main.py'),
                                             'that_nlp_library.model_main.ModelController.predict_ddict': ( 'model_main.html#modelcontroller.predict_ddict',
                                                                                                            'that_nlp_library/model_main.py'),
                                             'that_nlp_library.model_main.ModelController.predict_raw_dset': ( 'model_main.html#modelcontroller.predict_raw_dset',
                                                                                                               'that_nlp_library/model_main.py'),
                                             'that_nlp_library.model_main.ModelController.predict_raw_text': ( 'model_main.html#modelcontroller.predict_raw_text',
                                                                                                               'that_nlp_library/model_main.py'),
                                             'that_nlp_library.model_main._convert_pred_id_to_label': ( 'model_main.html#_convert_pred_id_to_label',
                                                                                                        'that_nlp_library/model_main.py'),
                                             'that_nlp_library.model_main._forward_pass_prediction': ( 'model_main.html#_forward_pass_prediction',
                                                                                                       'that_nlp_library/model_main.py'),
                                             'that_nlp_library.model_main.compute_metrics': ( 'model_main.html#compute_metrics',
                                                                                              'that_nlp_library/model_main.py'),
                                             'that_nlp_library.model_main.compute_metrics_separate_heads': ( 'model_main.html#compute_metrics_separate_heads',
                                                                                                             'that_nlp_library/model_main.py'),
                                             'that_nlp_library.model_main.finetune': ( 'model_main.html#finetune',
                                                                                       'that_nlp_library/model_main.py'),
                                             'that_nlp_library.model_main.loss_for_classification': ( 'model_main.html#loss_for_classification',
                                                                                                      'that_nlp_library/model_main.py'),
                                             'that_nlp_library.model_main.model_init_classification': ( 'model_main.html#model_init_classification',
                                                                                                        'that_nlp_library/model_main.py')},
            'that_nlp_library.models.gpt2.classifiers': { 'that_nlp_library.models.gpt2.classifiers.GPT2BaseForSequenceClassification': ( 'models.gpt2.classifiers.html#gpt2baseforsequenceclassification',
                                                                                                                                          'that_nlp_library/models/gpt2/classifiers.py'),
                                                          'that_nlp_library.models.gpt2.classifiers.GPT2BaseForSequenceClassification.__init__': ( 'models.gpt2.classifiers.html#gpt2baseforsequenceclassification.__init__',
                                                                                                                                                   'that_nlp_library/models/gpt2/classifiers.py'),
                                                          'that_nlp_library.models.gpt2.classifiers.GPT2BaseForSequenceClassification.forward': ( 'models.gpt2.classifiers.html#gpt2baseforsequenceclassification.forward',
                                                                                                                                                  'that_nlp_library/models/gpt2/classifiers.py'),
                                                          'that_nlp_library.models.gpt2.classifiers.GPT2HiddenStateConcatForSequenceClassification': ( 'models.gpt2.classifiers.html#gpt2hiddenstateconcatforsequenceclassification',
                                                                                                                                                       'that_nlp_library/models/gpt2/classifiers.py'),
                                                          'that_nlp_library.models.gpt2.classifiers.GPT2HiddenStateConcatForSequenceClassification.__init__': ( 'models.gpt2.classifiers.html#gpt2hiddenstateconcatforsequenceclassification.__init__',
                                                                                                                                                                'that_nlp_library/models/gpt2/classifiers.py'),
                                                          'that_nlp_library.models.gpt2.classifiers.GPT2HiddenStateConcatForSequenceClassification.forward': ( 'models.gpt2.classifiers.html#gpt2hiddenstateconcatforsequenceclassification.forward',
                                                                                                                                                               'that_nlp_library/models/gpt2/classifiers.py')},
            'that_nlp_library.models.roberta.classifiers': { 'that_nlp_library.models.roberta.classifiers.ConcatHeadExtended': ( 'models.roberta.classifiers.html#concatheadextended',
                                                                                                                                 'that_nlp_library/models/roberta/classifiers.py'),
                                                             'that_nlp_library.models.roberta.classifiers.ConcatHeadExtended.__init__': ( 'models.roberta.classifiers.html#concatheadextended.__init__',
                                                                                                                                          'that_nlp_library/models/roberta/classifiers.py'),
                                                             'that_nlp_library.models.roberta.classifiers.ConcatHeadExtended.forward': ( 'models.roberta.classifiers.html#concatheadextended.forward',
                                                                                                                                         'that_nlp_library/models/roberta/classifiers.py'),
                                                             'that_nlp_library.models.roberta.classifiers.ConcatHeadSimple': ( 'models.roberta.classifiers.html#concatheadsimple',
                                                                                                                               'that_nlp_library/models/roberta/classifiers.py'),
                                                             'that_nlp_library.models.roberta.classifiers.ConcatHeadSimple.__init__': ( 'models.roberta.classifiers.html#concatheadsimple.__init__',
                                                                                                                                        'that_nlp_library/models/roberta/classifiers.py'),
                                                             'that_nlp_library.models.roberta.classifiers.ConcatHeadSimple.forward': ( 'models.roberta.classifiers.html#concatheadsimple.forward',
                                                                                                                                       'that_nlp_library/models/roberta/classifiers.py'),
                                                             'that_nlp_library.models.roberta.classifiers.RobertaBaseForSequenceClassification': ( 'models.roberta.classifiers.html#robertabaseforsequenceclassification',
                                                                                                                                                   'that_nlp_library/models/roberta/classifiers.py'),
                                                             'that_nlp_library.models.roberta.classifiers.RobertaBaseForSequenceClassification.__init__': ( 'models.roberta.classifiers.html#robertabaseforsequenceclassification.__init__',
                                                                                                                                                            'that_nlp_library/models/roberta/classifiers.py'),
                                                             'that_nlp_library.models.roberta.classifiers.RobertaBaseForSequenceClassification.forward': ( 'models.roberta.classifiers.html#robertabaseforsequenceclassification.forward',
                                                                                                                                                           'that_nlp_library/models/roberta/classifiers.py'),
                                                             'that_nlp_library.models.roberta.classifiers.RobertaClassificationHeadCustom': ( 'models.roberta.classifiers.html#robertaclassificationheadcustom',
                                                                                                                                              'that_nlp_library/models/roberta/classifiers.py'),
                                                             'that_nlp_library.models.roberta.classifiers.RobertaClassificationHeadCustom.__init__': ( 'models.roberta.classifiers.html#robertaclassificationheadcustom.__init__',
                                                                                                                                                       'that_nlp_library/models/roberta/classifiers.py'),
                                                             'that_nlp_library.models.roberta.classifiers.RobertaClassificationHeadCustom.forward': ( 'models.roberta.classifiers.html#robertaclassificationheadcustom.forward',
                                                                                                                                                      'that_nlp_library/models/roberta/classifiers.py'),
                                                             'that_nlp_library.models.roberta.classifiers.RobertaHiddenStateConcatForSequenceClassification': ( 'models.roberta.classifiers.html#robertahiddenstateconcatforsequenceclassification',
                                                                                                                                                                'that_nlp_library/models/roberta/classifiers.py'),
                                                             'that_nlp_library.models.roberta.classifiers.RobertaHiddenStateConcatForSequenceClassification.__init__': ( 'models.roberta.classifiers.html#robertahiddenstateconcatforsequenceclassification.__init__',
                                                                                                                                                                         'that_nlp_library/models/roberta/classifiers.py'),
                                                             'that_nlp_library.models.roberta.classifiers.RobertaHiddenStateConcatForSequenceClassification.forward': ( 'models.roberta.classifiers.html#robertahiddenstateconcatforsequenceclassification.forward',
                                                                                                                                                                        'that_nlp_library/models/roberta/classifiers.py')},
            'that_nlp_library.models.roberta.conditional_prob_classifiers': { 'that_nlp_library.models.roberta.conditional_prob_classifiers.RobertaHSCCProbSequenceClassification': ( 'models.roberta.conditional_prob_classifiers.html#robertahsccprobsequenceclassification',
                                                                                                                                                                                      'that_nlp_library/models/roberta/conditional_prob_classifiers.py'),
                                                                              'that_nlp_library.models.roberta.conditional_prob_classifiers.RobertaHSCCProbSequenceClassification.__init__': ( 'models.roberta.conditional_prob_classifiers.html#robertahsccprobsequenceclassification.__init__',
                                                                                                                                                                                               'that_nlp_library/models/roberta/conditional_prob_classifiers.py'),
                                                                              'that_nlp_library.models.roberta.conditional_prob_classifiers.RobertaHSCCProbSequenceClassification.forward': ( 'models.roberta.conditional_prob_classifiers.html#robertahsccprobsequenceclassification.forward',
                                                                                                                                                                                              'that_nlp_library/models/roberta/conditional_prob_classifiers.py'),
                                                                              'that_nlp_library.models.roberta.conditional_prob_classifiers.build_standard_condition_mask': ( 'models.roberta.conditional_prob_classifiers.html#build_standard_condition_mask',
                                                                                                                                                                              'that_nlp_library/models/roberta/conditional_prob_classifiers.py')},
            'that_nlp_library.models.roberta.deep_hierarchical_classifiers': { 'that_nlp_library.models.roberta.deep_hierarchical_classifiers.RobertaConcatHeadDHCRoot': ( 'models.roberta.deep_hierarchical_classifiers.html#robertaconcatheaddhcroot',
                                                                                                                                                                           'that_nlp_library/models/roberta/deep_hierarchical_classifiers.py'),
                                                                               'that_nlp_library.models.roberta.deep_hierarchical_classifiers.RobertaConcatHeadDHCRoot.__init__': ( 'models.roberta.deep_hierarchical_classifiers.html#robertaconcatheaddhcroot.__init__',
                                                                                                                                                                                    'that_nlp_library/models/roberta/deep_hierarchical_classifiers.py'),
                                                                               'that_nlp_library.models.roberta.deep_hierarchical_classifiers.RobertaConcatHeadDHCRoot.forward': ( 'models.roberta.deep_hierarchical_classifiers.html#robertaconcatheaddhcroot.forward',
                                                                                                                                                                                   'that_nlp_library/models/roberta/deep_hierarchical_classifiers.py'),
                                                                               'that_nlp_library.models.roberta.deep_hierarchical_classifiers.RobertaHSCDHCSequenceClassification': ( 'models.roberta.deep_hierarchical_classifiers.html#robertahscdhcsequenceclassification',
                                                                                                                                                                                      'that_nlp_library/models/roberta/deep_hierarchical_classifiers.py'),
                                                                               'that_nlp_library.models.roberta.deep_hierarchical_classifiers.RobertaHSCDHCSequenceClassification.__init__': ( 'models.roberta.deep_hierarchical_classifiers.html#robertahscdhcsequenceclassification.__init__',
                                                                                                                                                                                               'that_nlp_library/models/roberta/deep_hierarchical_classifiers.py'),
                                                                               'that_nlp_library.models.roberta.deep_hierarchical_classifiers.RobertaHSCDHCSequenceClassification.forward': ( 'models.roberta.deep_hierarchical_classifiers.html#robertahscdhcsequenceclassification.forward',
                                                                                                                                                                                              'that_nlp_library/models/roberta/deep_hierarchical_classifiers.py'),
                                                                               'that_nlp_library.models.roberta.deep_hierarchical_classifiers.RobertaSimpleHSCDHCSequenceClassification': ( 'models.roberta.deep_hierarchical_classifiers.html#robertasimplehscdhcsequenceclassification',
                                                                                                                                                                                            'that_nlp_library/models/roberta/deep_hierarchical_classifiers.py'),
                                                                               'that_nlp_library.models.roberta.deep_hierarchical_classifiers.RobertaSimpleHSCDHCSequenceClassification.__init__': ( 'models.roberta.deep_hierarchical_classifiers.html#robertasimplehscdhcsequenceclassification.__init__',
                                                                                                                                                                                                     'that_nlp_library/models/roberta/deep_hierarchical_classifiers.py'),
                                                                               'that_nlp_library.models.roberta.deep_hierarchical_classifiers.RobertaSimpleHSCDHCSequenceClassification.forward': ( 'models.roberta.deep_hierarchical_classifiers.html#robertasimplehscdhcsequenceclassification.forward',
                                                                                                                                                                                                    'that_nlp_library/models/roberta/deep_hierarchical_classifiers.py'),
                                                                               'that_nlp_library.models.roberta.deep_hierarchical_classifiers._check_hierarchy': ( 'models.roberta.deep_hierarchical_classifiers.html#_check_hierarchy',
                                                                                                                                                                   'that_nlp_library/models/roberta/deep_hierarchical_classifiers.py'),
                                                                               'that_nlp_library.models.roberta.deep_hierarchical_classifiers.build_DHC_conditional_mask': ( 'models.roberta.deep_hierarchical_classifiers.html#build_dhc_conditional_mask',
                                                                                                                                                                             'that_nlp_library/models/roberta/deep_hierarchical_classifiers.py'),
                                                                               'that_nlp_library.models.roberta.deep_hierarchical_classifiers.loss_for_DHC': ( 'models.roberta.deep_hierarchical_classifiers.html#loss_for_dhc',
                                                                                                                                                               'that_nlp_library/models/roberta/deep_hierarchical_classifiers.py')},
            'that_nlp_library.text_augmentation': { 'that_nlp_library.text_augmentation.fill_mask_augmentation': ( 'text_augmentation.html#fill_mask_augmentation',
                                                                                                                   'that_nlp_library/text_augmentation.py'),
                                                    'that_nlp_library.text_augmentation.remove_vnmese_accent': ( 'text_augmentation.html#remove_vnmese_accent',
                                                                                                                 'that_nlp_library/text_augmentation.py')},
            'that_nlp_library.text_main': { 'that_nlp_library.text_main.TextDataController': ( 'text_main.html#textdatacontroller',
                                                                                               'that_nlp_library/text_main.py'),
                                            'that_nlp_library.text_main.TextDataController.__init__': ( 'text_main.html#textdatacontroller.__init__',
                                                                                                        'that_nlp_library/text_main.py'),
                                            'that_nlp_library.text_main.TextDataController._check_sup_types': ( 'text_main.html#textdatacontroller._check_sup_types',
                                                                                                                'that_nlp_library/text_main.py'),
                                            'that_nlp_library.text_main.TextDataController._check_validation_leaking': ( 'text_main.html#textdatacontroller._check_validation_leaking',
                                                                                                                         'that_nlp_library/text_main.py'),
                                            'that_nlp_library.text_main.TextDataController._convert_regression_to_float': ( 'text_main.html#textdatacontroller._convert_regression_to_float',
                                                                                                                            'that_nlp_library/text_main.py'),
                                            'that_nlp_library.text_main.TextDataController._create_label_mapping_func': ( 'text_main.html#textdatacontroller._create_label_mapping_func',
                                                                                                                          'that_nlp_library/text_main.py'),
                                            'that_nlp_library.text_main.TextDataController._determine_multihead_multilabel': ( 'text_main.html#textdatacontroller._determine_multihead_multilabel',
                                                                                                                               'that_nlp_library/text_main.py'),
                                            'that_nlp_library.text_main.TextDataController._determine_val_key': ( 'text_main.html#textdatacontroller._determine_val_key',
                                                                                                                  'that_nlp_library/text_main.py'),
                                            'that_nlp_library.text_main.TextDataController._do_augmentation': ( 'text_main.html#textdatacontroller._do_augmentation',
                                                                                                                'that_nlp_library/text_main.py'),
                                            'that_nlp_library.text_main.TextDataController._do_filtering': ( 'text_main.html#textdatacontroller._do_filtering',
                                                                                                             'that_nlp_library/text_main.py'),
                                            'that_nlp_library.text_main.TextDataController._do_label_transformation': ( 'text_main.html#textdatacontroller._do_label_transformation',
                                                                                                                        'that_nlp_library/text_main.py'),
                                            'that_nlp_library.text_main.TextDataController._do_train_shuffling': ( 'text_main.html#textdatacontroller._do_train_shuffling',
                                                                                                                   'that_nlp_library/text_main.py'),
                                            'that_nlp_library.text_main.TextDataController._do_transformation': ( 'text_main.html#textdatacontroller._do_transformation',
                                                                                                                  'that_nlp_library/text_main.py'),
                                            'that_nlp_library.text_main.TextDataController._encode_labels': ( 'text_main.html#textdatacontroller._encode_labels',
                                                                                                              'that_nlp_library/text_main.py'),
                                            'that_nlp_library.text_main.TextDataController._process_metadatas': ( 'text_main.html#textdatacontroller._process_metadatas',
                                                                                                                  'that_nlp_library/text_main.py'),
                                            'that_nlp_library.text_main.TextDataController._simplify_ddict': ( 'text_main.html#textdatacontroller._simplify_ddict',
                                                                                                               'that_nlp_library/text_main.py'),
                                            'that_nlp_library.text_main.TextDataController._train_test_split': ( 'text_main.html#textdatacontroller._train_test_split',
                                                                                                                 'that_nlp_library/text_main.py'),
                                            'that_nlp_library.text_main.TextDataController._upsampling': ( 'text_main.html#textdatacontroller._upsampling',
                                                                                                           'that_nlp_library/text_main.py'),
                                            'that_nlp_library.text_main.TextDataController.do_all_preprocessing': ( 'text_main.html#textdatacontroller.do_all_preprocessing',
                                                                                                                    'that_nlp_library/text_main.py'),
                                            'that_nlp_library.text_main.TextDataController.do_tokenization': ( 'text_main.html#textdatacontroller.do_tokenization',
                                                                                                               'that_nlp_library/text_main.py'),
                                            'that_nlp_library.text_main.TextDataController.from_csv': ( 'text_main.html#textdatacontroller.from_csv',
                                                                                                        'that_nlp_library/text_main.py'),
                                            'that_nlp_library.text_main.TextDataController.from_df': ( 'text_main.html#textdatacontroller.from_df',
                                                                                                       'that_nlp_library/text_main.py'),
                                            'that_nlp_library.text_main.TextDataController.from_pickle': ( 'text_main.html#textdatacontroller.from_pickle',
                                                                                                           'that_nlp_library/text_main.py'),
                                            'that_nlp_library.text_main.TextDataController.prepare_test_dataset': ( 'text_main.html#textdatacontroller.prepare_test_dataset',
                                                                                                                    'that_nlp_library/text_main.py'),
                                            'that_nlp_library.text_main.TextDataController.prepare_test_dataset_from_csv': ( 'text_main.html#textdatacontroller.prepare_test_dataset_from_csv',
                                                                                                                             'that_nlp_library/text_main.py'),
                                            'that_nlp_library.text_main.TextDataController.prepare_test_dataset_from_df': ( 'text_main.html#textdatacontroller.prepare_test_dataset_from_df',
                                                                                                                            'that_nlp_library/text_main.py'),
                                            'that_nlp_library.text_main.TextDataController.prepare_test_dataset_from_raws': ( 'text_main.html#textdatacontroller.prepare_test_dataset_from_raws',
                                                                                                                              'that_nlp_library/text_main.py'),
                                            'that_nlp_library.text_main.TextDataController.process_and_tokenize': ( 'text_main.html#textdatacontroller.process_and_tokenize',
                                                                                                                    'that_nlp_library/text_main.py'),
                                            'that_nlp_library.text_main.TextDataController.save_as_pickles': ( 'text_main.html#textdatacontroller.save_as_pickles',
                                                                                                               'that_nlp_library/text_main.py'),
                                            'that_nlp_library.text_main.TextDataController.set_data_collator': ( 'text_main.html#textdatacontroller.set_data_collator',
                                                                                                                 'that_nlp_library/text_main.py'),
                                            'that_nlp_library.text_main.TextDataController.set_verbose': ( 'text_main.html#textdatacontroller.set_verbose',
                                                                                                           'that_nlp_library/text_main.py'),
                                            'that_nlp_library.text_main.TextDataController.validate_input': ( 'text_main.html#textdatacontroller.validate_input',
                                                                                                              'that_nlp_library/text_main.py'),
                                            'that_nlp_library.text_main.concat_metadatas': ( 'text_main.html#concat_metadatas',
                                                                                             'that_nlp_library/text_main.py'),
                                            'that_nlp_library.text_main.tokenize_function': ( 'text_main.html#tokenize_function',
                                                                                              'that_nlp_library/text_main.py'),
                                            'that_nlp_library.text_main.tokenizer_explain': ( 'text_main.html#tokenizer_explain',
                                                                                              'that_nlp_library/text_main.py'),
                                            'that_nlp_library.text_main.two_steps_tokenization_explain': ( 'text_main.html#two_steps_tokenization_explain',
                                                                                                           'that_nlp_library/text_main.py')},
            'that_nlp_library.text_main_lm': { 'that_nlp_library.text_main_lm.TextDataLMController': ( 'text_main_lm.html#textdatalmcontroller',
                                                                                                       'that_nlp_library/text_main_lm.py'),
                                               'that_nlp_library.text_main_lm.TextDataLMController.__init__': ( 'text_main_lm.html#textdatalmcontroller.__init__',
                                                                                                                'that_nlp_library/text_main_lm.py'),
                                               'that_nlp_library.text_main_lm.TextDataLMController._do_augmentation': ( 'text_main_lm.html#textdatalmcontroller._do_augmentation',
                                                                                                                        'that_nlp_library/text_main_lm.py'),
                                               'that_nlp_library.text_main_lm.TextDataLMController._do_label_transformation': ( 'text_main_lm.html#textdatalmcontroller._do_label_transformation',
                                                                                                                                'that_nlp_library/text_main_lm.py'),
                                               'that_nlp_library.text_main_lm.TextDataLMController._do_train_shuffling': ( 'text_main_lm.html#textdatalmcontroller._do_train_shuffling',
                                                                                                                           'that_nlp_library/text_main_lm.py'),
                                               'that_nlp_library.text_main_lm.TextDataLMController._encode_labels': ( 'text_main_lm.html#textdatalmcontroller._encode_labels',
                                                                                                                      'that_nlp_library/text_main_lm.py'),
                                               'that_nlp_library.text_main_lm.TextDataLMController._group_texts_with_stride': ( 'text_main_lm.html#textdatalmcontroller._group_texts_with_stride',
                                                                                                                                'that_nlp_library/text_main_lm.py'),
                                               'that_nlp_library.text_main_lm.TextDataLMController._upsampling': ( 'text_main_lm.html#textdatalmcontroller._upsampling',
                                                                                                                   'that_nlp_library/text_main_lm.py'),
                                               'that_nlp_library.text_main_lm.TextDataLMController.do_all_preprocessing': ( 'text_main_lm.html#textdatalmcontroller.do_all_preprocessing',
                                                                                                                            'that_nlp_library/text_main_lm.py'),
                                               'that_nlp_library.text_main_lm.TextDataLMController.do_tokenization': ( 'text_main_lm.html#textdatalmcontroller.do_tokenization',
                                                                                                                       'that_nlp_library/text_main_lm.py'),
                                               'that_nlp_library.text_main_lm.TextDataLMController.prepare_test_dataset': ( 'text_main_lm.html#textdatalmcontroller.prepare_test_dataset',
                                                                                                                            'that_nlp_library/text_main_lm.py'),
                                               'that_nlp_library.text_main_lm.TextDataLMController.prepare_test_dataset_from_raws': ( 'text_main_lm.html#textdatalmcontroller.prepare_test_dataset_from_raws',
                                                                                                                                      'that_nlp_library/text_main_lm.py'),
                                               'that_nlp_library.text_main_lm.TextDataLMController.process_and_tokenize': ( 'text_main_lm.html#textdatalmcontroller.process_and_tokenize',
                                                                                                                            'that_nlp_library/text_main_lm.py'),
                                               'that_nlp_library.text_main_lm.TextDataLMController.save_as_pickles': ( 'text_main_lm.html#textdatalmcontroller.save_as_pickles',
                                                                                                                       'that_nlp_library/text_main_lm.py'),
                                               'that_nlp_library.text_main_lm.TextDataLMController.set_data_collator': ( 'text_main_lm.html#textdatalmcontroller.set_data_collator',
                                                                                                                         'that_nlp_library/text_main_lm.py')},
            'that_nlp_library.text_main_lm_streaming': { 'that_nlp_library.text_main_lm_streaming.TextDataLMControllerStreaming': ( 'text_main_lm_streaming.html#textdatalmcontrollerstreaming',
                                                                                                                                    'that_nlp_library/text_main_lm_streaming.py'),
                                                         'that_nlp_library.text_main_lm_streaming.TextDataLMControllerStreaming.__init__': ( 'text_main_lm_streaming.html#textdatalmcontrollerstreaming.__init__',
                                                                                                                                             'that_nlp_library/text_main_lm_streaming.py'),
                                                         'that_nlp_library.text_main_lm_streaming.TextDataLMControllerStreaming._construct_generator_with_batch': ( 'text_main_lm_streaming.html#textdatalmcontrollerstreaming._construct_generator_with_batch',
                                                                                                                                                                    'that_nlp_library/text_main_lm_streaming.py'),
                                                         'that_nlp_library.text_main_lm_streaming.TextDataLMControllerStreaming._do_label_transformation': ( 'text_main_lm_streaming.html#textdatalmcontrollerstreaming._do_label_transformation',
                                                                                                                                                             'that_nlp_library/text_main_lm_streaming.py'),
                                                         'that_nlp_library.text_main_lm_streaming.TextDataLMControllerStreaming._do_transformation_augmentation_tokenization': ( 'text_main_lm_streaming.html#textdatalmcontrollerstreaming._do_transformation_augmentation_tokenization',
                                                                                                                                                                                 'that_nlp_library/text_main_lm_streaming.py'),
                                                         'that_nlp_library.text_main_lm_streaming.TextDataLMControllerStreaming._do_transformation_tokenization': ( 'text_main_lm_streaming.html#textdatalmcontrollerstreaming._do_transformation_tokenization',
                                                                                                                                                                    'that_nlp_library/text_main_lm_streaming.py'),
                                                         'that_nlp_library.text_main_lm_streaming.TextDataLMControllerStreaming._do_transformation_tokenization_generator': ( 'text_main_lm_streaming.html#textdatalmcontrollerstreaming._do_transformation_tokenization_generator',
                                                                                                                                                                              'that_nlp_library/text_main_lm_streaming.py'),
                                                         'that_nlp_library.text_main_lm_streaming.TextDataLMControllerStreaming._do_transformation_tokenization_generator_fast': ( 'text_main_lm_streaming.html#textdatalmcontrollerstreaming._do_transformation_tokenization_generator_fast',
                                                                                                                                                                                   'that_nlp_library/text_main_lm_streaming.py'),
                                                         'that_nlp_library.text_main_lm_streaming.TextDataLMControllerStreaming._encode_labels': ( 'text_main_lm_streaming.html#textdatalmcontrollerstreaming._encode_labels',
                                                                                                                                                   'that_nlp_library/text_main_lm_streaming.py'),
                                                         'that_nlp_library.text_main_lm_streaming.TextDataLMControllerStreaming._group_texts_with_stride': ( 'text_main_lm_streaming.html#textdatalmcontrollerstreaming._group_texts_with_stride',
                                                                                                                                                             'that_nlp_library/text_main_lm_streaming.py'),
                                                         'that_nlp_library.text_main_lm_streaming.TextDataLMControllerStreaming.prepare_test_dataset': ( 'text_main_lm_streaming.html#textdatalmcontrollerstreaming.prepare_test_dataset',
                                                                                                                                                         'that_nlp_library/text_main_lm_streaming.py'),
                                                         'that_nlp_library.text_main_lm_streaming.TextDataLMControllerStreaming.prepare_test_dataset_from_raws': ( 'text_main_lm_streaming.html#textdatalmcontrollerstreaming.prepare_test_dataset_from_raws',
                                                                                                                                                                   'that_nlp_library/text_main_lm_streaming.py'),
                                                         'that_nlp_library.text_main_lm_streaming.TextDataLMControllerStreaming.process_and_tokenize': ( 'text_main_lm_streaming.html#textdatalmcontrollerstreaming.process_and_tokenize',
                                                                                                                                                         'that_nlp_library/text_main_lm_streaming.py'),
                                                         'that_nlp_library.text_main_lm_streaming.TextDataLMControllerStreaming.save_as_pickles': ( 'text_main_lm_streaming.html#textdatalmcontrollerstreaming.save_as_pickles',
                                                                                                                                                    'that_nlp_library/text_main_lm_streaming.py'),
                                                         'that_nlp_library.text_main_lm_streaming.TextDataLMControllerStreaming.set_data_collator': ( 'text_main_lm_streaming.html#textdatalmcontrollerstreaming.set_data_collator',
                                                                                                                                                      'that_nlp_library/text_main_lm_streaming.py')},
            'that_nlp_library.text_main_streaming': { 'that_nlp_library.text_main_streaming.TextDataControllerStreaming': ( 'text_main_streaming.html#textdatacontrollerstreaming',
                                                                                                                            'that_nlp_library/text_main_streaming.py'),
                                                      'that_nlp_library.text_main_streaming.TextDataControllerStreaming.__init__': ( 'text_main_streaming.html#textdatacontrollerstreaming.__init__',
                                                                                                                                     'that_nlp_library/text_main_streaming.py'),
                                                      'that_nlp_library.text_main_streaming.TextDataControllerStreaming._check_sup_types': ( 'text_main_streaming.html#textdatacontrollerstreaming._check_sup_types',
                                                                                                                                             'that_nlp_library/text_main_streaming.py'),
                                                      'that_nlp_library.text_main_streaming.TextDataControllerStreaming._construct_generator_with_batch': ( 'text_main_streaming.html#textdatacontrollerstreaming._construct_generator_with_batch',
                                                                                                                                                            'that_nlp_library/text_main_streaming.py'),
                                                      'that_nlp_library.text_main_streaming.TextDataControllerStreaming._convert_regression_to_float': ( 'text_main_streaming.html#textdatacontrollerstreaming._convert_regression_to_float',
                                                                                                                                                         'that_nlp_library/text_main_streaming.py'),
                                                      'that_nlp_library.text_main_streaming.TextDataControllerStreaming._create_label_mapping_func': ( 'text_main_streaming.html#textdatacontrollerstreaming._create_label_mapping_func',
                                                                                                                                                       'that_nlp_library/text_main_streaming.py'),
                                                      'that_nlp_library.text_main_streaming.TextDataControllerStreaming._determine_multihead_multilabel': ( 'text_main_streaming.html#textdatacontrollerstreaming._determine_multihead_multilabel',
                                                                                                                                                            'that_nlp_library/text_main_streaming.py'),
                                                      'that_nlp_library.text_main_streaming.TextDataControllerStreaming._do_filtering': ( 'text_main_streaming.html#textdatacontrollerstreaming._do_filtering',
                                                                                                                                          'that_nlp_library/text_main_streaming.py'),
                                                      'that_nlp_library.text_main_streaming.TextDataControllerStreaming._do_label_transformation': ( 'text_main_streaming.html#textdatacontrollerstreaming._do_label_transformation',
                                                                                                                                                     'that_nlp_library/text_main_streaming.py'),
                                                      'that_nlp_library.text_main_streaming.TextDataControllerStreaming._do_transformation_augmentation_tokenization': ( 'text_main_streaming.html#textdatacontrollerstreaming._do_transformation_augmentation_tokenization',
                                                                                                                                                                         'that_nlp_library/text_main_streaming.py'),
                                                      'that_nlp_library.text_main_streaming.TextDataControllerStreaming._do_transformation_augmentation_tokenization_generator': ( 'text_main_streaming.html#textdatacontrollerstreaming._do_transformation_augmentation_tokenization_generator',
                                                                                                                                                                                   'that_nlp_library/text_main_streaming.py'),
                                                      'that_nlp_library.text_main_streaming.TextDataControllerStreaming._do_transformation_augmentation_tokenization_generator_linebyline': ( 'text_main_streaming.html#textdatacontrollerstreaming._do_transformation_augmentation_tokenization_generator_linebyline',
                                                                                                                                                                                              'that_nlp_library/text_main_streaming.py'),
                                                      'that_nlp_library.text_main_streaming.TextDataControllerStreaming._do_transformation_tokenization': ( 'text_main_streaming.html#textdatacontrollerstreaming._do_transformation_tokenization',
                                                                                                                                                            'that_nlp_library/text_main_streaming.py'),
                                                      'that_nlp_library.text_main_streaming.TextDataControllerStreaming._encode_labels': ( 'text_main_streaming.html#textdatacontrollerstreaming._encode_labels',
                                                                                                                                           'that_nlp_library/text_main_streaming.py'),
                                                      'that_nlp_library.text_main_streaming.TextDataControllerStreaming._process_metadatas': ( 'text_main_streaming.html#textdatacontrollerstreaming._process_metadatas',
                                                                                                                                               'that_nlp_library/text_main_streaming.py'),
                                                      'that_nlp_library.text_main_streaming.TextDataControllerStreaming._simplify_ddict': ( 'text_main_streaming.html#textdatacontrollerstreaming._simplify_ddict',
                                                                                                                                            'that_nlp_library/text_main_streaming.py'),
                                                      'that_nlp_library.text_main_streaming.TextDataControllerStreaming.from_pickle': ( 'text_main_streaming.html#textdatacontrollerstreaming.from_pickle',
                                                                                                                                        'that_nlp_library/text_main_streaming.py'),
                                                      'that_nlp_library.text_main_streaming.TextDataControllerStreaming.prepare_test_dataset': ( 'text_main_streaming.html#textdatacontrollerstreaming.prepare_test_dataset',
                                                                                                                                                 'that_nlp_library/text_main_streaming.py'),
                                                      'that_nlp_library.text_main_streaming.TextDataControllerStreaming.prepare_test_dataset_from_csv': ( 'text_main_streaming.html#textdatacontrollerstreaming.prepare_test_dataset_from_csv',
                                                                                                                                                          'that_nlp_library/text_main_streaming.py'),
                                                      'that_nlp_library.text_main_streaming.TextDataControllerStreaming.prepare_test_dataset_from_df': ( 'text_main_streaming.html#textdatacontrollerstreaming.prepare_test_dataset_from_df',
                                                                                                                                                         'that_nlp_library/text_main_streaming.py'),
                                                      'that_nlp_library.text_main_streaming.TextDataControllerStreaming.prepare_test_dataset_from_raws': ( 'text_main_streaming.html#textdatacontrollerstreaming.prepare_test_dataset_from_raws',
                                                                                                                                                           'that_nlp_library/text_main_streaming.py'),
                                                      'that_nlp_library.text_main_streaming.TextDataControllerStreaming.process_and_tokenize': ( 'text_main_streaming.html#textdatacontrollerstreaming.process_and_tokenize',
                                                                                                                                                 'that_nlp_library/text_main_streaming.py'),
                                                      'that_nlp_library.text_main_streaming.TextDataControllerStreaming.save_as_pickles': ( 'text_main_streaming.html#textdatacontrollerstreaming.save_as_pickles',
                                                                                                                                            'that_nlp_library/text_main_streaming.py'),
                                                      'that_nlp_library.text_main_streaming.TextDataControllerStreaming.set_data_collator': ( 'text_main_streaming.html#textdatacontrollerstreaming.set_data_collator',
                                                                                                                                              'that_nlp_library/text_main_streaming.py'),
                                                      'that_nlp_library.text_main_streaming.TextDataControllerStreaming.set_verbose': ( 'text_main_streaming.html#textdatacontrollerstreaming.set_verbose',
                                                                                                                                        'that_nlp_library/text_main_streaming.py')},
            'that_nlp_library.text_transformation': { 'that_nlp_library.text_transformation.apply_vnmese_word_tokenize': ( 'text_transformation.html#apply_vnmese_word_tokenize',
                                                                                                                           'that_nlp_library/text_transformation.py')},
            'that_nlp_library.utils': { 'that_nlp_library.utils.HiddenPrints': ('utils.html#hiddenprints', 'that_nlp_library/utils.py'),
                                        'that_nlp_library.utils.HiddenPrints.__enter__': ( 'utils.html#hiddenprints.__enter__',
                                                                                           'that_nlp_library/utils.py'),
                                        'that_nlp_library.utils.HiddenPrints.__exit__': ( 'utils.html#hiddenprints.__exit__',
                                                                                          'that_nlp_library/utils.py'),
                                        'that_nlp_library.utils.augmentation_stream_generator': ( 'utils.html#augmentation_stream_generator',
                                                                                                  'that_nlp_library/utils.py'),
                                        'that_nlp_library.utils.callable_name': ('utils.html#callable_name', 'that_nlp_library/utils.py'),
                                        'that_nlp_library.utils.check_and_get_attribute': ( 'utils.html#check_and_get_attribute',
                                                                                            'that_nlp_library/utils.py'),
                                        'that_nlp_library.utils.check_input_validation': ( 'utils.html#check_input_validation',
                                                                                           'that_nlp_library/utils.py'),
                                        'that_nlp_library.utils.check_text_leaking': ( 'utils.html#check_text_leaking',
                                                                                       'that_nlp_library/utils.py'),
                                        'that_nlp_library.utils.create_dir': ('utils.html#create_dir', 'that_nlp_library/utils.py'),
                                        'that_nlp_library.utils.func_all': ('utils.html#func_all', 'that_nlp_library/utils.py'),
                                        'that_nlp_library.utils.get_dset_col_names': ( 'utils.html#get_dset_col_names',
                                                                                       'that_nlp_library/utils.py'),
                                        'that_nlp_library.utils.hf_filter_dset': ('utils.html#hf_filter_dset', 'that_nlp_library/utils.py'),
                                        'that_nlp_library.utils.hf_map_dset': ('utils.html#hf_map_dset', 'that_nlp_library/utils.py'),
                                        'that_nlp_library.utils.lambda_batch': ('utils.html#lambda_batch', 'that_nlp_library/utils.py'),
                                        'that_nlp_library.utils.lambda_map_batch': ( 'utils.html#lambda_map_batch',
                                                                                     'that_nlp_library/utils.py'),
                                        'that_nlp_library.utils.load_pickle': ('utils.html#load_pickle', 'that_nlp_library/utils.py'),
                                        'that_nlp_library.utils.none2emptystr': ('utils.html#none2emptystr', 'that_nlp_library/utils.py'),
                                        'that_nlp_library.utils.print_msg': ('utils.html#print_msg', 'that_nlp_library/utils.py'),
                                        'that_nlp_library.utils.resize_model_embeddings': ( 'utils.html#resize_model_embeddings',
                                                                                            'that_nlp_library/utils.py'),
                                        'that_nlp_library.utils.save_to_pickle': ('utils.html#save_to_pickle', 'that_nlp_library/utils.py'),
                                        'that_nlp_library.utils.seed_everything': ( 'utils.html#seed_everything',
                                                                                    'that_nlp_library/utils.py'),
                                        'that_nlp_library.utils.seed_notorch': ('utils.html#seed_notorch', 'that_nlp_library/utils.py'),
                                        'that_nlp_library.utils.sigmoid': ('utils.html#sigmoid', 'that_nlp_library/utils.py'),
                                        'that_nlp_library.utils.val2iterable': ('utils.html#val2iterable', 'that_nlp_library/utils.py')}}}



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/08_evaluations.ipynb.

# %% ../nbs/08_evaluations.ipynb 25
from __future__ import annotations
from .utils import *
from sklearn.metrics import classification_report,ConfusionMatrixDisplay
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt


# %% auto 0
__all__ = ['evaluate_classification_model_metadata', 'show_top_n_predictions', 'evaluate_classification_model']

# %% ../nbs/08_evaluations.ipynb 26
def _show_metrics(label,pred,metric_funcs):
    results=''
    for m_func in metric_funcs:
        m_name=callable_name(m_func)
        results+=f'{m_name}: {m_func(label,pred):.4f} . '
    return results

# %% ../nbs/08_evaluations.ipynb 27
def evaluate_classification_model_metadata(df:pd.DataFrame, # The main dataframe containing the predictions
                                           metadatas:str|list, # Metadata(s) to perform analysis
                                           label_name:str, # Label's column name
                                           pred_name:str, # Prediction's column name
                                           metric_funcs:list, # Metric(s) to calculate
                                          ):
    metadatas = val2iterable(metadatas)
    metric_funcs = val2iterable(metric_funcs)
    
    for metadata in metadatas:
        print_msg(f'{label_name} Analysis on metadata: {metadata}',20)
        print_msg('Distribution',10)
        print(pd.concat((df[metadata].value_counts(),df[metadata].value_counts(normalize=True)),axis=1))
        print()
        print_msg(f'Metrics for each value in {metadata}',10)
        print(f'- For all data:\n{_show_metrics(df[label_name].values,df[pred_name].values,metric_funcs)}')
        for val in np.sort(df[metadata].unique()):
            y_true=df.loc[df[metadata]==val,label_name].values
            y_pred=df.loc[df[metadata]==val,pred_name].values
            print(f'- For {val}:\n{_show_metrics(y_true,y_pred,metric_funcs)}')

# %% ../nbs/08_evaluations.ipynb 31
def show_top_n_predictions(df:pd.DataFrame, # The main dataframe containing the predictions
                          text_name:str, # Text's column name
                          label_name:str, # Label's column name
                          pred_name:str, # Prediction's column name
                          prob_name:str, # Prediction probability's column name
                          is_incorrect=True, # To show top correct or incorrect sentences
                          ascending=False, # To sort by prob_name ascendingly or descendingly
                          n_show=10, # Number of sentences to show
                         ):
    if is_incorrect:
        df = df[df[pred_name]!=df[label_name]]
    else:
        df = df[df[pred_name]==df[label_name]]
    df = df.sort_values(prob_name,ascending=ascending)[[text_name,label_name,pred_name,prob_name]].head(n_show).values
    for row in df:
        print(f'Text: {row[0]}')
        print(f'True label: {row[1]}, {"but" if is_incorrect else ""} predict {row[2]}, with confidence {row[3]:.4f}')
        print('-'*60)

# %% ../nbs/08_evaluations.ipynb 35
def evaluate_classification_model(df:pd.DataFrame, # The main dataframe containing the predictions
                              text_name:str, # Text's column name
                              label_name:str, # Label's column name
                              pred_name:str, # Prediction's column name
                              prob_name:str, # Prediction probability's column name
                              metric_funcs:list=[], # Metric(s) to calculate
                              metadatas:str|list=[], # Metadata(s) to perform analysis
                              n_show=10, # Number of sentences to show
                              cm_figsize=(20,20), # Confusion matrix's figure size
                             ):

    y_true = df[label_name].values
    y_pred = df[pred_name].values
    print_msg(f'{label_name} Analysis',20)
    print()
    print_msg('Classification Report',15)
    print(classification_report(y_true,y_pred))
    print()
    print_msg('Confusion Matrix',15)
    fig,ax = plt.subplots(figsize=cm_figsize)
    ConfusionMatrixDisplay.from_predictions(y_true,y_pred,xticks_rotation='vertical',ax=ax)
    plt.show()
    print()
    evaluate_classification_model_metadata(df,metadatas,label_name,pred_name,metric_funcs)
    print()
    print_msg("Let's look at some wrong predictions with high confidence",15)
    show_top_n_predictions(df,text_name,label_name,pred_name,prob_name,
                           is_incorrect=True,
                           ascending=False,
                           n_show=n_show)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/02_text_transformation.ipynb.

# %% ../nbs/02_text_transformation.ipynb 3
from __future__ import annotations
from underthesea import word_tokenize, sent_tokenize, text_normalize

# %% auto 0
__all__ = ['apply_vnmese_word_tokenize']

# %% ../nbs/02_text_transformation.ipynb 4
def apply_vnmese_word_tokenize(sentence:str, # Input sentence
                        normalize_text=False, # To 'normalize' the text before tokenization
                        fixed_words=[]
                       ):
    "Applying UnderTheSea Vietnamese word tokenization"
    if normalize_text:
        sentence = text_normalize(sentence)
    sens = sent_tokenize(sentence)

    tokenized_sen = []
    for sen in sens:
        tokenized_sen.append(word_tokenize(sen,format='text',fixed_words=fixed_words))
    return ' '.join(tokenized_sen)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/01_text_augmentation.ipynb.

# %% ../nbs/01_text_augmentation.ipynb 3
from __future__ import annotations
import unidecode
import random

# %% auto 0
__all__ = ['remove_vnmese_accent', 'fill_mask_augmentation']

# %% ../nbs/01_text_augmentation.ipynb 4
def remove_vnmese_accent(sentence:str, # Input sentence
                         prob=1, # Probability that this function is applied to the text
                        ):
    "Perform Vietnamese accent removal"
    return unidecode.unidecode(sentence) if random.random()<prob else sentence

# %% ../nbs/01_text_augmentation.ipynb 9
def fill_mask_augmentation(sentence:str, # Input Sentence,
                           fillmask_pipeline, # HuggingFace fill-mask pipeline
                           prob=1, # Probability that this function is applied to the text
                           random_top_k=1, # To select output randomly from top k mask filled
                          ):
    # References: https://huggingface.co/docs/datasets/v2.14.1/en/process#data-augmentation
    if random.random()>=prob: return sentence
    mask_token = fillmask_pipeline.tokenizer.mask_token
    words = sentence.split(' ')
    K = random.randint(1, len(words)-1)
    masked_sentence = " ".join(words[:K]  + [mask_token] + words[K+1:])
    predictions = fillmask_pipeline(masked_sentence,top_k = random_top_k)
    weights = [p['score'] for p in predictions]
    sentences = [p['sequence'] for p in predictions]
    return random.choices(sentences, weights = weights, k = 1)[0]
    



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_text_main_lm.ipynb.

# %% ../nbs/00_text_main_lm.ipynb 3
from __future__ import annotations
from datasets import Dataset
from .utils import *
from .text_main import *
from functools import partial
import warnings
from transformers import DataCollatorForLanguageModeling

# %% auto 0
__all__ = ['TextDataLMController']

# %% ../nbs/00_text_main_lm.ipynb 6
class TextDataLMController(TextDataController):
    def __init__(self,
                 inp, # HuggingFainpce Dataset or DatasetDict
                 main_text:str, # Name of the main text column
                 filter_dict={}, # A dictionary: {feature: filtering_function_for_that_feature}
                 metadatas=[], # Names of the metadata columns
                 process_metas=True, # Whether to do simple text processing on the chosen metadatas
                 content_transformations=[], # A list of text transformations
                 val_ratio:int|float|None=0.2, # Ratio of data for validation set
                 stratify_cols=[], # Column(s) needed to do stratified shuffle split
                 seed=None, # Random seed
                 batch_size=1024, # CPU batch size
                 num_proc=4, # Number of process for multiprocessing
                 cols_to_keep=None, # Columns to keep after all processings
                 verbose=True, # Whether to prdint processing information
                ):
        super().__init__(inp=inp,
                         main_text=main_text,
                         filter_dict=filter_dict,
                         metadatas=metadatas,
                         process_metas=process_metas,
                         content_transformations=content_transformations,
                         val_ratio=val_ratio,
                         stratify_cols=stratify_cols,
                         seed=seed,
                         batch_size=batch_size,
                         num_proc=num_proc,
                         cols_to_keep=cols_to_keep,
                         verbose=verbose
                        )
            
    
    def _do_label_transformation(self):
        raise NotImplementedError("There's no classification/regression label in text processing for Language Model")
        
    def _encode_labels(self):
        raise NotImplementedError("There's no classification/regression label in text processing for Language Model")

    
    def _upsampling(self):
        raise NotImplementedError("There's no upsampling in text processing for Language Model")
      
    def _do_augmentation(self):
        raise NotImplementedError("There's no text augmentation in text processing for Language Model")
     
    def save_as_pickles(self,
                        fname, # Name of the pickle file
                        parent='pickle_files', # Parent folder
                       ):
        
        save_to_pickle(self,fname,parent=parent) 
        
    def _do_train_shuffling(self):
        print_msg('Shuffling and flattening train set',20,verbose=self.verbose)
        self.main_ddict['train'] = self.main_ddict['train'].shuffle(seed=self.seed).flatten_indices(num_proc = self.num_proc)
        self.verboseprint('Done')

    def _group_texts_with_stride(self,examples):
        max_length = self.max_length
        if max_length is None: 
            max_length = self.tokenizer.model_max_length
        stride = self.stride
        if stride is None: stride=max_length
        else: stride = max_length-stride
        if stride==0: raise ValueError(f'Stride cannot be equal to max length of {max_length}')
            
        # Concatenate all texts.
        concatenated_examples = {k: sum(examples[k], []) for k in examples.keys()}
        total_length = len(concatenated_examples[list(examples.keys())[0]])
        result_all={}
        for k,t in concatenated_examples.items():
            result=[]
            i=0
            while i+max_length<=total_length:
                result.append(t[i:i+max_length])
                i+=stride
            result_all[k]=result

        return result_all  
    

    def do_all_preprocessing(self,
                             shuffle_trn=True # To shuffle the train set before tokenization
                            ):
        if self._processed_call:
            warnings.warn('Your dataset has already been processed. Returning the previous processed DatasetDict...')
            return self.main_ddict
            
        print_msg('Start Main Text Processing',20,verbose=self.verbose)
        
        # Filtering
        self.dset,self.ddict_rest = self._do_filtering(self.dset,self.ddict_rest)
        
        # Process metadatas
        self.dset,self.ddict_rest = self._process_metadatas(self.dset,self.ddict_rest)
        
        
        # Content transformation
        self.dset,self.ddict_rest = self._do_transformation(self.dset,self.ddict_rest)
         
        # Train Test Split.
        ### self.main_ddict is created here
        self._train_test_split()
        
        # Dropping unused columns
        self._simplify_ddict()
        
        # Check validation leaking
        self._check_validation_leaking()
        
        # Shuffle train
        if shuffle_trn:
            self._do_train_shuffling()
        
        self._processed_call=True
        
        return self.main_ddict
    
        
    def do_tokenization(self,
                        tokenizer, # Tokenizer (preferably from HuggingFace)
                        max_length=None, # pad to model's allowed max length (default is max_sequence_length). Use -1 for no padding at all
                        line_by_line=True, # To whether tokenize each sentence separately, or concatenate them
                        stride=None, # option to do striding when line_by_line is False
                        trn_size=None, # The number of training data to be tokenized
                        tok_num_proc=None, # Number of processes for tokenization
                       ):
        # References
#         https://github.com/huggingface/transformers/blob/main/examples/pytorch/language-modeling/run_mlm.py
#         https://github.com/huggingface/transformers/blob/main/examples/pytorch/language-modeling/run_clm.py
        
        print_msg('Tokenization',20,verbose=self.verbose)
        self.tokenizer = tokenizer
        self.max_length = max_length
        self.line_by_line = line_by_line
        self.stride = stride
        self.tok_num_proc = tok_num_proc if tok_num_proc else self.num_proc
        
        tok_func = partial(tokenize_function,tok=self.tokenizer,
                           max_length=max_length if line_by_line else -1,
                           return_special_tokens_mask=True)
        
        _func = partial(lambda_map_batch,
                        feature=self.main_text,
                        func=tok_func,
                        output_feature=None,
                        is_batched=self.is_batched)
        
        
        if trn_size is not None:
            if isinstance(trn_size,float):
                num_shard = int(1/trn_size)
            else: # int
                trn_len=len(self.main_ddict['train'])
                num_shard = trn_len//trn_size
            self.main_ddict['train'] = self.main_ddict['train'].shard(num_shard,0)
        
        for k in self.main_ddict.keys():
            self.main_ddict[k] = hf_map_dset(self.main_ddict[k],_func,self.is_batched,self.batch_size,self.tok_num_proc)
            self.main_ddict[k] = self.main_ddict[k].remove_columns(self.cols_to_keep)
        
        if not line_by_line: # token concatenation
            for k in self.main_ddict.keys():
                self.main_ddict[k] = hf_map_dset(self.main_ddict[k],
                                                 self._group_texts_with_stride,
                                                 is_batched=True,
                                                 batch_size=self.batch_size if self.batch_size>1 else 1024,
                                                 num_proc=self.tok_num_proc)
                
        
        self.verboseprint('Done')
        return self.main_ddict
        
    def process_and_tokenize(self,
                             tokenizer, # Tokenizer (preferably from HuggingFace)
                             max_length=None, # pad to model's allowed max length (default is max_sequence_length)
                             line_by_line=True, # To whether tokenize each sentence separately, or concatenate them and then tokenize
                             stride=None, # option to do striding when line_by_line is False
                             trn_size=None, # The number of training data to be tokenized
                             tok_num_proc=None, # Number of processes for tokenization
                             shuffle_trn=True, # To shuffle the train set before tokenization
                            ):
        """
        This will perform `do_all_processing` then `do_tokenization`
        """
        if self.seed:
            seed_everything(self.seed)
        _ = self.do_all_preprocessing(shuffle_trn)
        _ = self.do_tokenization(tokenizer,max_length,line_by_line,stride,trn_size,tok_num_proc)
        
    
    def set_data_collator(self,
                          is_mlm=True, # Is this masked language model (True) or causal language model (False)
                          mlm_prob=0.15, # Mask probability for masked language model
                         ):
        if not hasattr(self,'max_length'):
            raise ValueError("Please call `process_and_tokenize' or `do_tokenization` to tokenize your dataset")
        
        self.is_mlm = is_mlm
        pad_to_multiple_of_8 = (self.max_length<0) # get data collator to pad
        self.data_collator = DataCollatorForLanguageModeling(tokenizer=self.tokenizer,
                                                             mlm=is_mlm,
                                                             mlm_probability=mlm_prob,
                                                             pad_to_multiple_of=8 if pad_to_multiple_of_8 else None
                                                            )
                                               
    
    def prepare_test_dataset_from_raws(self,
                                       content, # Either a single sentence, list of sentence or a dictionary with keys are metadata columns and values are list
                                       do_tokenize=False, # Whether to tokenize text
                                      ):
        if len(self.metadatas) and not isinstance(content,dict):
            raise ValueError(f'There is/are metadatas in the preprocessing step. Please include a dictionary including these keys for metadatas: {self.metadatas}, and texture content: {self.main_text}')
            
        _dic = {self.main_text:[content]} if isinstance(content,str) else content
        for k in _dic.keys():
            _dic[k] = val2iterable(_dic[k])
        
        test_dict = Dataset.from_dict(_dic)
        
        # set num_proc to 1 for small data processing
        _tmp1 = self.num_proc
        _tmp2 = self.tok_num_proc
        self.num_proc=1
        self.tok_num_proc=1
        results = self.prepare_test_dataset(test_dict,do_tokenize)
        self.num_proc = _tmp1
        self.tok_num_proc=_tmp2
        return results
        
    def prepare_test_dataset(self,
                             test_dset, # The HuggingFace Dataset as Test set
                             do_tokenize, # Whether to tokenize text
                            ):
        test_cols = set(get_dset_col_names(test_dset))
        missing_cols = set(self.cols_to_keep) - test_cols
        if len(missing_cols):
            raise ValueError(f'Test set does not have these columns required for preprocessings: {missing_cols}')
            
        print_msg('Start Test Set Transformation',20,verbose=self.verbose)
        
        # Process metadatas
        test_dset = self._process_metadatas(test_dset)
        
        # Content transformation
        test_dset = self._do_transformation(test_dset)
        
        if not do_tokenize:
            # Drop every columns except for main_text
            cols_to_remove = {c for c in test_cols if c!=self.main_text}
            test_dset = test_dset.remove_columns(list(cols_to_remove))
        else:
            # Drop unused columns
            cols_to_remove = test_cols - set(self.cols_to_keep)
            test_dset = test_dset.remove_columns(list(cols_to_remove))
            
            print_msg('Tokenization',20,verbose=self.verbose)
            tok_func = partial(tokenize_function,
                           tok=self.tokenizer,
                           max_length=self.max_length if self.line_by_line else -1,
                           return_special_tokens_mask=True
                          )
            
            _func = partial(lambda_map_batch,
                        feature=self.main_text,
                        func=tok_func,
                        output_feature=None,
                        is_batched=self.is_batched)
            test_dset = hf_map_dset(test_dset,_func,self.is_batched,self.batch_size,self.tok_num_proc)
            
        self.verboseprint('Done')
        return test_dset



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/03_model_main.ipynb.

# %% ../nbs/03_model_main.ipynb 3
from __future__ import annotations
import os
from transformers import Trainer, TrainingArguments, AutoConfig
from datasets import DatasetDict,Dataset
import torch
import gc
from sklearn.metrics import accuracy_score
from functools import partial
import numpy as np
from .utils import *
from .text_main import TextDataController
from .text_main_streaming import TextDataControllerStreaming

# %% auto 0
__all__ = ['model_init_classification', 'compute_metrics', 'compute_metrics_separate_heads', 'loss_for_classification',
           'finetune', 'ModelController']

# %% ../nbs/03_model_main.ipynb 4
def model_init_classification(
                              model_class, # Model's class object, e.g. RobertaHiddenStateConcatForSequenceClassification
                              cpoint_path, # Either model string name on HuggingFace, or the path to model checkpoint
                              output_hidden_states:bool, # To whether output the model hidden states or not. Useful when you try to build a custom classification head 
                              device=None, # Device to train on
                              config=None, # Model config. If not provided, AutoConfig is used to load config from cpoint_path
                              seed=None, # Random seed
                              body_model=None, # If not none, we use this to initialize model's body. If you only want to load the model checkpoint in cpoint_path, leave this as none
                              model_kwargs={} # Keyword arguments for model (both head and body)
                             ):
    """To initialize a classification (or regression) model, either from an existing HuggingFace model or custom architecture
    
    Can be used for binary, multi-class single-head, multi-class multi-head, multi-label clasisifcation, and regression
    """
    if device is None: device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

    if config is None:
        config = AutoConfig.from_pretrained(
            cpoint_path,
            output_hidden_states=output_hidden_states,
        )
    else:
        config.output_hidden_states=output_hidden_states
    
    if seed:
        seed_everything(seed)
        
    if body_model is not None:
        model = model_class(config=config,**model_kwargs)
#         layers = list(model.children())
        print('Loading body weights. This assumes the body is the very first block of your custom architecture')
        body_name, _ = next(iter(model.named_children()))
        setattr(model, body_name, body_model)
        model = model.to(device)
        
    else:
        model = model_class.from_pretrained(cpoint_path,config=config,**model_kwargs).to(device)
    
    print(f'Total parameters: {sum(param.numel() for param in model.parameters())}')
    print(f'Total trainable parameters: {sum(param.numel() for param in model.parameters() if param.requires_grad)}')
    
    return model

# %% ../nbs/03_model_main.ipynb 7
def compute_metrics(pred, # An EvalPrediction object from HuggingFace (which is a named tuple with ```predictions``` and ```label_ids``` attributes)
                    metric_funcs=[], # A list of metric functions to evaluate
                    metric_types=[], # Type of metric ('classification' or 'regression') for each metric functions above
                    head_sizes=[], # Class size for each head. Regression head will have head size 1
                    label_names=[], # Names of the label (dependent variable) columns
                    is_multilabel=False, # Whether this is a multilabel classification
                    multilabel_threshold=0.5 # Threshold for multilabel (>= threshold is positive)
                   ):
    """
    Return a dictionary of metric name and its values.
    
    Reference: https://github.com/huggingface/transformers/blob/main/src/transformers/trainer_utils.py#L107C16-L107C16
    """
    assert len(head_sizes)==len(label_names)
    labels = pred.label_ids 
    if isinstance(pred.predictions,tuple):
        preds = pred.predictions[0]
    else:
        preds = pred.predictions
        
    results={}
    metric_funcs = val2iterable(metric_funcs)
    metric_types = val2iterable(metric_types)
    if len(metric_types):
        assert len(metric_funcs)==len(metric_types),"The length of `metric_types` must equal to the length of `metric_funcs`"
    for i,(_size,_name) in enumerate(zip(head_sizes,label_names)):
        sup_type='classification'
        start= 0 if i==0 else start+head_sizes[i-1]
        end = start + _size
        _pred = preds[:,start:end]
        if is_multilabel:
            # sigmoid and threshold
            _pred = (sigmoid(_pred)>=multilabel_threshold).astype(int)
        elif _size>1: # classification
            _pred = _pred.argmax(-1)
        else: # _size==1, regression
            sup_type='regression'
        
        # set up default metric_types, assuming all heads are doing the same type
        if len(metric_types)==0:
            metric_types = val2iterable(sup_type,len(metric_funcs))
        
        _label = labels[:,i] if len(head_sizes)>1 else labels
        for i,m_func in enumerate(metric_funcs):
            if metric_types[i]==sup_type:
                m_name = callable_name(m_func)
                results[f'{m_name}_{_name}']=m_func(_label,_pred)
    return results

# %% ../nbs/03_model_main.ipynb 9
def compute_metrics_separate_heads(pred, # An EvalPrediction object from HuggingFace (which is a named tuple with ```predictions``` and ```label_ids``` attributes)
                              metric_funcs=[], # A list of metric functions to evaluate
                              label_names=[], # Names of the label (dependent variable) columns
                              **kwargs
                             ):
    """
    Return a dictionary of metric name and its values. This is used in Deep Hierarchical Classification (special case of multi-head classification)
    
    This metric function is mainly used when you have a separate logit output for each head 
    (instead of the typical multi-head logit output: all heads' logits are concatenated)
    """
    # pred: EvalPrediction object 
    # (which is a named tuple with predictions and label_ids attributes)
    labels = pred.label_ids # (bs,number of head separately)
    assert labels.shape[1]==len(label_names)
    
    results={}
    metric_funcs = val2iterable(metric_funcs)
    
    for i in range(len(label_names)):
        _label = labels[:,i]
        _pred = pred.predictions[i].argmax(-1)
        for m_func in metric_funcs:
            m_name = callable_name(m_func)
            results[f'{m_name}_{label_names[i]}']=m_func(_label,_pred)
    
    return results

# %% ../nbs/03_model_main.ipynb 11
def loss_for_classification(logits, # output of the last linear layer, before any softmax/sigmoid. Size: (bs,class_size)
                            labels, # determined by your datasetdict. Size: (bs,number_of_head)
                            is_multilabel=False, # Whether this is a multilabel classification
                            is_multihead=False, # Whether this is a multihead classification
                            head_sizes=[], # Class size for each head. Regression head will have head size 1
                            head_weights=[], # loss weight for each head. Default to 1 for each head
                           ):
    """
    The general loss function for classification
    
    - If is_multilabel is ```False``` and is_multihead is ```False```: Single-Head Classification, e.g. You predict 1 out of n class
    
    - If is_multilabel is ```False``` and is_multihead is ```True```: Multi-Head Classification, e.g. You predict 1 out of n classes at Level 1, 
    and 1 out of m classes at Level 2
    
    - If is_multilabel is ```True``` and is_multihead is ```False```: Single-Head Multi-Label Classification, e.g. You predict x out of n class (x>=0)
    
    - If is_multilabel is ```True``` and is_multihead is ```True```: Not supported
    
    """
    if is_multilabel and is_multihead: raise ValueError('Multi-Label and Multi-Head problem is not supported')
    head_sizes = val2iterable(head_sizes)
    if len(head_sizes) and not len(head_weights):
        head_weights = val2iterable(1,len(head_sizes))
        
    loss=0
    if not is_multilabel:
        assert len(head_sizes)==len(head_weights),"Make sure len of head_sizes and len of head_weights are equal"
        for i,(_size,_weight) in enumerate(zip(head_sizes,head_weights)):
            start= 0 if i==0 else start+head_sizes[i-1]
            end = start + _size
            
            loss_fct = torch.nn.MSELoss() if _size==1 else torch.nn.CrossEntropyLoss()
            
            _logits = logits[:,start:end] # (bs, _size)
            _logits = _logits.squeeze() if _size ==1 else _logits.view(-1,_size)
            
            _label = labels[:,i] if len(head_sizes)>1 else labels # (bs,) for 1 total 1 head size, (bs,num_head) otherwise
            _label = _label.view(-1)
            if isinstance(loss_fct,torch.nn.CrossEntropyLoss): _label = _label.long()
            loss = loss + _weight*loss_fct(_logits,_label)
    else:
        loss_fct = torch.nn.BCEWithLogitsLoss()
        loss = loss_fct(logits,
                        labels.float())
                    
    return loss

# %% ../nbs/03_model_main.ipynb 13
def finetune(lr, # Learning rate
             bs, # Batch size
             wd, # Weight decay
             epochs, # Number of epochs
             ddict, # The HuggingFace datasetdict
             tokenizer,# HuggingFace tokenizer
             o_dir = './tmp_weights', # Directory to save weights
             save_checkpoint=False, # Whether to save weights (checkpoints) to o_dir
             model=None, # NLP model
             model_init=None, # A function to initialize model
             data_collator=None, # HuggingFace data collator
             compute_metrics=None, # A function to compute metric, e.g. `compute_metrics`
             grad_accum_steps=2, # The batch at each step will be divided by this integer and gradient will be accumulated over gradient_accumulation_steps steps.
             lr_scheduler_type='cosine',  # The scheduler type to use. Including: linear, cosine, cosine_with_restarts, polynomial, constant, constant_with_warmup
             warmup_ratio=0.1, # The warmup ratio for some lr scheduler
             no_valid=False, # Whether there is a validation set or not
             seed=None, # Random seed
             report_to='none', # The list of integrations to report the results and logs to. Supported platforms are "azure_ml", "comet_ml", "mlflow", "neptune", "tensorboard","clearml" and "wandb". Use "all" to report to all integrations installed, "none" for no integrations.
             trainer_class=None, # You can include the class name of your custom trainer here
            ):
    "The main model training/finetuning function"
    torch.cuda.empty_cache()
    gc.collect()
    
    if seed:
        seed_everything(seed)
        
    training_args = TrainingArguments(o_dir, 
                                learning_rate=lr, 
                                warmup_ratio=warmup_ratio,
                                lr_scheduler_type=lr_scheduler_type, 
                                fp16=True,
                                do_train=True,
                                do_eval= not no_valid,
                                evaluation_strategy="no" if no_valid else "epoch", 
                                save_strategy="epoch" if save_checkpoint else 'no',
                                overwrite_output_dir=True,
                                gradient_accumulation_steps=grad_accum_steps,
                                per_device_train_batch_size=bs, 
                                per_device_eval_batch_size=bs,
                                num_train_epochs=epochs, weight_decay=wd,
                                report_to=report_to,
                                logging_dir=os.path.join(o_dir, 'log') if report_to!='none' else None,
                                logging_steps = len(ddict["train"]) // bs,
                                )

    # instantiate trainer
    trainer_class = Trainer if trainer_class is None else trainer_class
    trainer = trainer_class(
        model=model,
        model_init=model_init if model is None else None,
        args=training_args,
        train_dataset=ddict['train'],#.shard(200, 0)
        eval_dataset=ddict['validation'] if not no_valid else None,
        data_collator=data_collator,
        tokenizer=tokenizer,
        compute_metrics=compute_metrics if not no_valid else None
    )
    
    trainer.train()
    return trainer

# %% ../nbs/03_model_main.ipynb 16
def _forward_pass_prediction(batch,
                             model=None, # NLP model
                             topk=1, # Number of labels to return for each head
                             is_multilabel=False, # Is this a multilabel classification?
                             multilabel_threshold=0.5, # The threshold for multilabel classification
                             model_input_names=['input_ids', 'token_type_ids', 'attention_mask'], # Model required inputs, from tokenizer.model_input_names
                             data_collator=None, # HuggingFace data collator
                             label_names=[], # Names of the label columns
                             head_sizes=[], # Class size for each head. Regression head will have head size 1
                             device = None, # device that the model is trained on
                             are_heads_separated=False, # is this multi-head, but each head has a separated logit?
                             ):
    if data_collator is not None:
        
# --- Convert from  
# {'input_ids': [tensor([    0, 10444,   244, 14585,   125,  2948,  5925,   368,     2]), 
#                tensor([    0, 16098,  2913,   244,   135,   198, 34629,  6356,     2])]
# 'attention_mask': [tensor([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]), 
#                    tensor([1, 1, 1, 1, 1, 1, 1, 1, 1])]
#                    }
# --- to
# [{'input_ids': tensor([    0, 10444,   244, 14585,   125,  2948,  5925,   368,     2]),
#   'attention_mask': tensor([1, 1, 1, 1, 1, 1, 1, 1, 1])},
#  {'input_ids': tensor([    0, 16098,  2913,   244,   135,   198, 34629,  6356,     2]),
#   'attention_mask': tensor([1, 1, 1, 1, 1, 1, 1, 1, 1])}]

        # remove string text, due to transformer new version       
        collator_inp = []
        ks = [k for k in batch.keys() if k in model_input_names+['label']]
        vs = [batch[k] for k in ks]
        for pair in zip(*vs):
            collator_inp.append({k:v for k,v in zip(ks,pair)})
        
        batch = data_collator(collator_inp)
    
    inputs = {k:v.to(device) for k,v in batch.items()
              if k in model_input_names}
    
    assert len(label_names)==len(head_sizes), "Length of `label_names` must equal to length of `head_sizes`"
        
    # switch to eval mode for evaluation
    if model.training:
        model.eval()
    with torch.no_grad():
        outputs = model(**inputs)
        outputs_logits = outputs.logits
        results={}
        for i in range(len(label_names)):
            if are_heads_separated:
                _f = partial(torch.nn.functional.softmax,dim=1)
                _preds = _f(outputs_logits[i].cpu())
            else:
                outputs_logits = outputs_logits.cpu()
                start = 0 if i==0 else start+head_sizes[i-1]
                end = start + head_sizes[i]
                if head_sizes[i]==1: # regression
                    _f = lambda x: x.squeeze()
                else: # multilabel, or classification
                    _f = partial(torch.nn.functional.softmax,dim=1) if not is_multilabel else torch.sigmoid 
                _preds = _f(outputs_logits[:,start:end])
        
            if is_multilabel:
                _pred_labels = _preds>=multilabel_threshold
                results[f'pred_{label_names[i]}']=_pred_labels.numpy()
                results[f'pred_prob_{label_names[i]}']=_preds.numpy()
            else:
                if head_sizes[i]==1: # regression
                    _result = _preds.numpy()
                    if _result.ndim==0: _result=_result.reshape((1,))
                    results[f'pred_{label_names[i]}']=_result
                else: # classification
                    _p,_l = torch.topk(_preds,topk,dim=-1)
                    if topk==1:
                        _l,_p = _l[:,0],_p[:,0]
                    results[f'pred_{label_names[i]}']=_l.numpy()
                    results[f'pred_prob_{label_names[i]}']=_p.numpy()    

    
    # Switch back to train mode
    if not model.training:
        model.train()
    
    return results

# %% ../nbs/03_model_main.ipynb 17
def _convert_pred_id_to_label(dset,
                              label_names,
                              label_lists,
                              topk=1,
                              is_multilabel=False,
                              batch_size=1000,
                              num_proc=1
                             ):
    is_batched=batch_size>1
    if is_multilabel:
        get_label_str_multilabel = lambda x: [label_lists[0][int(j)] for j in np.where(x==True)[0]]
        _func = partial(lambda_map_batch,feature=f'pred_{label_names[0]}',
                        func=get_label_str_multilabel,
                        is_batched=is_batched
                       )
        dset = hf_map_dset(dset,_func,
                           is_batched=is_batched,
                           batch_size=batch_size,
                           num_proc=num_proc
                          )
        return dset
    
    for i in range(len(label_names)):
        if len(label_lists[i])==0: # regression
            _func1 = lambda xs: xs
        else: # classification
            _func1 = lambda xs: label_lists[i][int(xs)] if not isinstance(xs,np.ndarray) else [label_lists[i][int(x)] for x in xs]
        
        _func2 = partial(lambda_map_batch,
                         feature=f'pred_{label_names[i]}',
                        func=_func1,
                        is_batched=is_batched
                       )
        dset = hf_map_dset(dset,_func2,
                           is_batched=is_batched,
                           batch_size=batch_size,
                           num_proc=num_proc
                           )
    return dset


# %% ../nbs/03_model_main.ipynb 18
class ModelController():
    def __init__(self,
                 model, # NLP model
                 data_store=None, # a TextDataController/TextDataControllerStreaming object
                 seed=None, # Random seed
                ):
        self.model = model
        self.data_store = data_store
        self.seed = seed
        
    def fit(self,
            epochs, # Number of epochs
            learning_rate, # Learning rate
            ddict=None, # DatasetDict to fit (will override data_store)
            metric_funcs=[accuracy_score], # A list of metric functions (can be from Sklearn)
            metric_types=[], # A list of metric types (`classification` or `regression`) that matches with the metric function list
            batch_size=16, # Batch size
            weight_decay=0.01, # Weight decay
            lr_scheduler_type='cosine', # The scheduler type to use. Including: linear, cosine, cosine_with_restarts, polynomial, constant, constant_with_warmup
            warmup_ratio=0.1, # The warmup ratio for some lr scheduler
            o_dir = './tmp_weights', # Directory to save weights
            save_checkpoint=False, # Whether to save weights (checkpoints) to o_dir
            hf_report_to='none', # The list of HuggingFace-allowed integrations to report the results and logs to
            compute_metrics=None, # A function to compute metric, e.g. `compute_metrics` which utilizes the given ```metric_funcs``` 
            grad_accum_steps=2, # Gradient will be accumulated over gradient_accumulation_steps steps.
            tokenizer=None, # Tokenizer (to override one in ```data_store```)
            data_collator=None, # Data Collator (to override one in ```data_store```)
            label_names=None, # Names of the label (dependent variable) columns (to override one in ```data_store```)
            head_sizes=None, # Class size for each head (to override one in ```model```)
            trainer_class=None, # You can include the class name of your custom trainer here
           ):
        
        if tokenizer is None: tokenizer=check_and_get_attribute(self.data_store,'tokenizer')
        if data_collator is None: data_collator=getattr(self.data_store,'data_collator',None)
        if ddict is None: ddict = check_and_get_attribute(self.data_store,'main_ddict')
            
        if label_names is None: label_names=check_and_get_attribute(self.data_store,'label_names')
        label_names = val2iterable(label_names)
        
        if head_sizes is None: 
            head_sizes=check_and_get_attribute(self.data_store,'label_lists')
            head_sizes = [len(hs) if len(hs) else 1 for hs in head_sizes]
        head_sizes = val2iterable(head_sizes)
        
        if len(set(ddict.keys()) & set(['train','training']))==0:
            raise ValueError("Missing the following key for DatasetDict: train/training")
        no_valid = len(set(ddict.keys()) & set(['validation','val','valid']))==0

        _compute_metrics = partial(compute_metrics,
                                   metric_funcs=metric_funcs,
                                   metric_types=metric_types,
                                   head_sizes=head_sizes,
                                   label_names=label_names 
                                  )
        
        trainer = finetune(learning_rate,batch_size,weight_decay,epochs,
                           ddict,tokenizer,o_dir,
                           save_checkpoint=save_checkpoint,
                           model=self.model,
                           data_collator=data_collator,
                           compute_metrics=_compute_metrics,
                           grad_accum_steps=grad_accum_steps,
                           lr_scheduler_type=lr_scheduler_type,
                           warmup_ratio=warmup_ratio,
                           no_valid=no_valid,
                           seed=self.seed,
                           trainer_class=trainer_class,
                           report_to=hf_report_to)
        self.trainer = trainer
        
    def predict_raw_text(self,
                         content:dict|list|str, # Either a single sentence, list of sentence or a dictionary where keys are metadata, values are list
                         is_multilabel=None, # Is this a multilabel classification?
                         multilabel_threshold=0.5, # Threshold for multilabel classification
                         topk=1, # Number of labels to return for each head
                         are_heads_separated=False # Are outpuf (of model) separate heads?
                        ):
        if not isinstance(self.data_store,(TextDataController,TextDataControllerStreaming)) or not self.data_store._processed_call:
            raise ValueError('This functionality needs a TextDataController object which has processed some training data')
#         with HiddenPrints():
        test_dset = self.data_store.prepare_test_dataset_from_raws(content)

        results = self.predict_ddict(ddict=test_dset,
                                     batch_size=1,
                                     is_multilabel=is_multilabel,
                                     multilabel_threshold=multilabel_threshold,
                                     topk=topk,
                                     are_heads_separated=are_heads_separated
                                    )
        return results.to_dict()
    
    def predict_raw_dset(self,
                         dset, # A raw HuggingFace dataset
                         batch_size=16, # GPU batch size. For a small amount of texts, you might want to keep this small
                         do_filtering=False, # Whether to perform data filtering on this test set
                         is_multilabel=None, # Is this a multilabel classification?
                         multilabel_threshold=0.5, # Threshold for multilabel classification
                         topk=1, # Number of labels to return for each head
                         are_heads_separated=False # Are outpuf (of model) separate heads?
                        ):
        if not isinstance(self.data_store,(TextDataController,TextDataControllerStreaming)) or not self.data_store._processed_call:
            raise ValueError('This functionality needs a TextDataController object which has processed some training data')
#         with HiddenPrints():
        test_dset = self.data_store.prepare_test_dataset(dset,do_filtering)

        results = self.predict_ddict(test_dset,
                                     batch_size=batch_size,
                                     is_multilabel=is_multilabel,
                                     multilabel_threshold=multilabel_threshold,
                                     topk=topk,
                                     are_heads_separated=are_heads_separated
                                    )
        return results
    
    def predict_ddict(self,
                      ddict:DatasetDict|Dataset=None, # A processed and tokenized DatasetDict/Dataset (will override one in ```data_store```)
                      ds_type='test', # The split of DatasetDict to predict
                      batch_size=16, # Batch size for making prediction on GPU
                      is_multilabel=None, # Is this a multilabel classification?
                      multilabel_threshold=0.5, # Threshold for multilabel classification
                      topk=1, # Number of labels to return for each head
                      tokenizer=None, # Tokenizer (to override one in ```data_store```)
                      data_collator=None, # Data Collator (to override one in ```data_store```)
                      label_names=None, # Names of the label (dependent variable) columns (to override one in ```data_store```)
                      class_names_predefined=None, # List of names associated with the labels (same index order) (to override one in ```data_store```)
                      are_heads_separated=False # Are outputs (of model) separate heads?
                     ):
        device = self.model.device
        if is_multilabel is None: is_multilabel=getattr(self.model,'is_multilabel',False)
        label_lists = class_names_predefined
        if tokenizer is None: tokenizer=check_and_get_attribute(self.data_store,'tokenizer')
        if data_collator is None: data_collator=getattr(self.data_store,'data_collator',None)
        if label_names is None: label_names=check_and_get_attribute(self.data_store,'label_names')
        if label_lists is None: label_lists = check_and_get_attribute(self.data_store,'label_lists')
        label_names = val2iterable(label_names)
        if not isinstance(label_lists[0],list):
            label_lists=[label_lists]    
        head_sizes = [len(hs) if len(hs) else 1 for hs in label_lists]
        if ddict is None: ddict = check_and_get_attribute(self.data_store,'main_ddict')
        if isinstance(ddict,DatasetDict):
            if ds_type not in ddict.keys():
                raise ValueError(f'{ds_type} is not in the given DatasetDict')
            ddict = ddict[ds_type]
            
        ddict.set_format("torch",
                        columns=tokenizer.model_input_names)
        
        print_msg('Start making predictions',20)
        # this will create features: pred_classname and/or pred_prob_classname
        results = ddict.map(
                            partial(_forward_pass_prediction,
                                    model=self.model,
                                    topk=topk,
                                    is_multilabel=is_multilabel,
                                    multilabel_threshold=multilabel_threshold,
                                    model_input_names=tokenizer.model_input_names,
                                    data_collator=data_collator,
                                    label_names=label_names,
                                    head_sizes=head_sizes,
                                    are_heads_separated = are_heads_separated,
                                    device=device
                                   ), 
                            batched=True, 
                            batch_size=batch_size)
        
        # convert all to numpy
        results.set_format('numpy')
            
        results = _convert_pred_id_to_label(results,
                                            label_names,
                                            label_lists,
                                            topk,
                                            is_multilabel,
                                            batch_size=1000,
                                            num_proc=1
                                           )
        return results
        



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_text_main.ipynb.

# %% ../nbs/00_text_main.ipynb 3
from __future__ import annotations
from sklearn.preprocessing import LabelEncoder,MultiLabelBinarizer
from datasets import DatasetDict,Dataset,IterableDataset,load_dataset,concatenate_datasets,Value
from pathlib import Path
from .utils import *
from functools import partial
import warnings
from datasets.utils.logging import disable_progress_bar,enable_progress_bar

# %% auto 0
__all__ = ['tokenizer_explain', 'two_steps_tokenization_explain', 'tokenize_function', 'concat_metadatas', 'TextDataController']

# %% ../nbs/00_text_main.ipynb 6
def tokenizer_explain(inp, # Input sentence
                      tokenizer, # Tokenizer (preferably from HuggingFace)
                      split_word=False # Is input `inp` split into list or not
                     ):
    "Display results from tokenizer"
    print('\t\t------- Tokenizer Explained -------')
    print('----- Input -----')
    print(inp)
    print()
    print('----- Tokenized results ----- ')
    print(tokenizer(inp,is_split_into_words=split_word))
    print()
    tok = tokenizer.encode(inp,is_split_into_words=split_word)
    print('----- Results from tokenizer.convert_ids_to_tokens -----')
    print(tokenizer.convert_ids_to_tokens(tok))
    print()
    print('----- Results from tokenizer.decode ----- ')
    print(tokenizer.decode(tok))
    print()

# %% ../nbs/00_text_main.ipynb 20
def two_steps_tokenization_explain(inp, # Input sentence
                                   tokenizer, # Tokenizer (preferably from HuggingFace)
                                   content_tfms=[], # A list of text transformations
                                   aug_tfms=[], # A list of text augmentation 
                                  ):
    "Display results form each content transformation, then display results from tokenizer"
    print('\t\t------- Text Transformation Explained -------')
    print('----- Raw sentence -----')
    print(inp)
    print()
    print('----- Content Transformations (on both train and test) -----')
    content_tfms = val2iterable(content_tfms)
    for tfm in content_tfms:
        print_msg(callable_name(tfm),3)
        inp = tfm(inp)
        print(inp)
        print()
    print()
    print('----- Augmentations (on train only) -----')
    aug_tfms = val2iterable(aug_tfms)
    for tfm in aug_tfms:
        print_msg(callable_name(tfm),3)
        inp = tfm(inp)
        print(inp)
        print()
    print()
    tokenizer_explain(inp,tokenizer)

# %% ../nbs/00_text_main.ipynb 41
def tokenize_function(text,
                      tok,
                      max_length=None,
                      is_split_into_words=False,
                      return_tensors=None,
                      return_special_tokens_mask=False
                     ):
    if max_length is None:
        # pad to model's default max sequence length
        return tok(text, padding="max_length", 
                   truncation=True,
                   is_split_into_words=is_split_into_words,
                   return_tensors=return_tensors,
                   return_special_tokens_mask=return_special_tokens_mask
                  )
    if isinstance(max_length,int) and max_length>0:
        # pad to the largest length of the current batch, and start truncating at max_length
        return tok(text, padding=True, 
                   max_length=max_length,
                   truncation=True,
                   is_split_into_words=is_split_into_words,
                   return_tensors=return_tensors,
                   return_special_tokens_mask=return_special_tokens_mask
                  )
    
    # no padding (still truncate at model's default max sequence length)
    return tok(text, padding=False,
               truncation=True,
               is_split_into_words=is_split_into_words,
               return_tensors=return_tensors,
               return_special_tokens_mask=return_special_tokens_mask
              )

# %% ../nbs/00_text_main.ipynb 56
def concat_metadatas(dset:dict, # HuggingFace Dataset
                     main_text, # Text feature name
                     metadatas, # Metadata (or a list of metadatas)
                     process_metas=True, # Whether apply simple metadata processing, i.e. space strip and lowercase
                     sep='.', # separator for contatenating to main_text
                     is_batched=True, # whether batching is applied
                    ):
    """
    Extract, process (optional) and concatenate metadatas to the front of text
    """
    results={main_text:dset[main_text]}
    for m in metadatas:
        m_data = dset[m]
        if process_metas:
            # just strip and lowercase
            m_data = [none2emptystr(v).strip().lower() for v in m_data] if is_batched else none2emptystr(m_data).strip().lower()
        results[m]=m_data
        if is_batched:
            results[main_text] = [f'{m_data[i]} {sep} {results[main_text][i]}' for i in range(len(m_data))]
        else:
            results[main_text] = f'{m_data} {sep} {results[main_text]}'
    return results

# %% ../nbs/00_text_main.ipynb 59
class TextDataController():
    def __init__(self,
                 inp, # HuggingFainpce Dataset or DatasetDict
                 main_text:str, # Name of the main text column
                 label_names=[], # Names of the label (dependent variable) columns
                 sup_types=[], # Type of supervised learning for each label name ('classification' or 'regression')
                 class_names_predefined=[], # List of names associated with the labels (same index order). Use empty list for regression
                 filter_dict={}, # A dictionary: {feature: filtering_function_for_that_feature}
                 label_tfm_dict={}, # A dictionary: {label_name: transform_function_for_that_label}
                 metadatas=[], # Names of the metadata columns
                 process_metas=True, # Whether to do simple text processing on the chosen metadatas
                 content_transformations=[], # A list of text transformations
                 val_ratio:int|float|None=0.2, # Ratio of data for validation set
                 stratify_cols=[], # Column(s) needed to do stratified shuffle split
                 upsampling_list={}, # A list of tuple. Each tuple: (feature,upsampling_function_based_on_the_feature)
                 content_augmentations=[], # A list of text augmentations
                 seed=None, # Random seed
                 batch_size=1024, # CPU batch size
                 num_proc=4, # Number of processes for multiprocessing
                 cols_to_keep=None, # Columns to keep after all processings
                 verbose=True, # Whether to prdint processing information
                ):
            
        self.main_text = main_text
        
        self.label_names = val2iterable(label_names)
        self.sup_types = val2iterable(sup_types)
        self._check_sup_types()
        self.label_lists = class_names_predefined
        
        self.filter_dict = filter_dict
        self.label_tfm_dict = label_tfm_dict
        self.metadatas = val2iterable(metadatas)
        self.process_metas = process_metas
        
        self.content_tfms = val2iterable(content_transformations)
        
        self.val_ratio = val_ratio
        self.stratify_cols = val2iterable(stratify_cols)
        
        self.upsampling_list = upsampling_list
        self.aug_tfms = val2iterable(content_augmentations)
        
        self.seed = seed
        self.is_batched = batch_size>1
        self.batch_size = batch_size
        self.num_proc = num_proc
        self.cols_to_keep = cols_to_keep
        
        self.ddict_rest = DatasetDict()
        self.verbose = verbose
        self.verboseprint = print if verbose else lambda *a, **k: None
        if not self.verbose:
            disable_progress_bar() # turn off huggingface `map` progress bar
        else:
            enable_progress_bar()
        if hasattr(inp,'keys'): # is datasetdict
            if 'train' in inp.keys(): 
                self.ddict_rest = inp
                self.dset = self.ddict_rest.pop('train')
            else:
                raise ValueError('The given DatasetDict has no "train" split')
        else: # is dataset
            self.dset = inp
        
        if isinstance(self.dset,IterableDataset):
            raise ValueError('TextDataController does not handle Iterable Dataset as input. Consider `TextDataControllerStreaming`')
        
        self._determine_val_key()
        self.all_cols = get_dset_col_names(self.dset)
        self._determine_multihead_multilabel()
        self._convert_regression_to_float()
        
        self._processed_call=False
            
    @classmethod
    def from_csv(cls,file_path,**kwargs):
        file_path = Path(file_path)
        ds = load_dataset(str(file_path.parent),
                                  data_files=file_path.name,
                                  split='train')
        return TextDataController(ds,**kwargs)
        
    
    @classmethod
    def from_df(cls,df,validate=True,**kwargs):
        if validate:
            check_input_validation(df)
        ds = Dataset.from_pandas(df)
        return TextDataController(ds,**kwargs)
    
    @classmethod
    def from_pickle(cls,
                    fname, # Name of the pickle file
                    parent='pickle_files' # Parent folder
                   ):
        return load_pickle(fname,parent=parent)

    def set_verbose(self,verbose):
        self.verbose = verbose
        self.verboseprint = print if verbose else lambda *a, **k: None
    
    def _convert_regression_to_float(self):
        if len(self.sup_types)==0: return
        # convert regression labels to float64
        reg_idxs = [i for i,v in enumerate(self.sup_types) if v=='regression']
        for i in reg_idxs:
            self.dset = self.dset.cast_column(self.label_names[i],Value("float64"))
            if self.val_key is not None:
                self.ddict_rest[self.val_key] = self.ddict_rest[self.val_key].cast_column(self.label_names[i],Value("float64"))
        
    def _check_sup_types(self):
        assert len(self.label_names)==len(self.sup_types), "The number of supervised learning declaration must equal to the number of label"
        assert len(set(self.sup_types) - set(['classification','regression']))==0, 'Accepted inputs for `sup_types` are `classification` and `regression`'
    
    def _determine_val_key(self):
        # determine validation split key
        self.val_key=None
        val_key = list(set(self.ddict_rest.keys()) & set(['val','validation','valid']))
        if len(val_key)>1: raise ValueError('Your DatasetDict has more than 1 validation split')
        if len(val_key)==1:
            val_key=val_key[0]
            self.val_key=val_key
            
    def _determine_multihead_multilabel(self):
        self.is_multilabel=False
        self.is_multihead=False
        if len(self.label_names)==0: return
        
        if len(self.label_names)>1:
            self.is_multihead=True
            
        # get label of first row
        first_label = self.dset[self.label_names[0]][0]
        if isinstance(first_label,(list,set,tuple)):
            # This is multi-label. Ignore self.label_names[1:]
            self.label_names = [self.label_names[0]]
            self.is_multihead=False
            self.is_multilabel=True
                     
    def validate_input(self):
        _df = self.dset.to_pandas()
        check_input_validation(_df)
    
    def save_as_pickles(self,
                        fname, # Name of the pickle file
                        parent='pickle_files', # Parent folder
                        drop_attributes=False # Whether to drop large-size attributes
                       ):
        if drop_attributes:
            if hasattr(self, 'main_ddict'):
                del self.main_ddict
            if hasattr(self, 'ddict_rest'):
                del self.ddict_rest
            if hasattr(self, 'aug_tfms'):
                del self.aug_tfms
        save_to_pickle(self,fname,parent=parent)
    
        
    def _check_validation_leaking(self):
        if self.val_ratio is None:
            return
        
        trn_txt = self.main_ddict['train'][self.main_text]
        val_txt = self.main_ddict['validation'][self.main_text]        
        val_txt_leaked = check_text_leaking(trn_txt,val_txt,verbose=self.verbose)
        
        if len(val_txt_leaked)==0: return
        
        # filter train dataset to get rid of leaks
        self.verboseprint('Filtering leaked data out of training set...')
        _func = partial(lambda_batch,
                        feature=self.main_text,
                        func=lambda x: x.strip().lower() not in val_txt_leaked,
                        is_batched=self.is_batched)
        self.main_ddict['train'] = hf_filter_dset(self.main_ddict['train'],_func,self.is_batched,self.batch_size,self.num_proc)   
        self.verboseprint('Done')
    
    def _process_metadatas(self,dset,ddict_rest=None):
        if len(self.metadatas):
            print_msg('Metadata Simple Processing & Concatenating to Main Content',verbose=self.verbose)
            map_func = partial(concat_metadatas,
                               main_text=self.main_text,
                               metadatas=self.metadatas,
                               process_metas=self.process_metas,
                               is_batched=self.is_batched)
            dset = hf_map_dset(dset,map_func,self.is_batched,self.batch_size,self.num_proc)
            if ddict_rest is not None:
                ddict_rest = hf_map_dset(ddict_rest,map_func,self.is_batched,self.batch_size,self.num_proc)
            self.verboseprint('Done')
        return dset if ddict_rest is None else (dset,ddict_rest)
    
    def _do_label_transformation(self):
        if len(self.label_names)==0 or len(self.label_tfm_dict)==0: return
        print_msg('Label Transformation',20,verbose=self.verbose)
        for f,tfm in self.label_tfm_dict.items():
            if f in self.label_names:
                _func = partial(lambda_map_batch,
                                feature=f,
                                func=tfm,
                                is_batched=self.is_batched
                               )                
                self.dset = hf_map_dset(self.dset,_func,self.is_batched,self.batch_size,self.num_proc)
                if self.val_key is not None:
                    self.ddict_rest[self.val_key] = hf_map_dset(self.ddict_rest[self.val_key],
                                                                _func,
                                                                self.is_batched,
                                                                self.batch_size,
                                                                self.num_proc)
        self.verboseprint('Done')
        
    def _create_label_mapping_func(self,encoder_classes):
        if self.is_multihead:
            label2idxs = [{v:i for i,v in enumerate(l_classes)} for l_classes in encoder_classes]
            _func = lambda inp: {'label': [[label2idxs[i][v] if len(label2idxs[i]) else v for i,v in enumerate(vs)] \
                                           for vs in zip(*[inp[l] for l in self.label_names])] if self.is_batched \
                                 else [label2idxs[i][v] if len(label2idxs[i]) else v for i,v in enumerate([inp[l] for l in self.label_names])]
                                }
            
        else: # single-head
            if self.sup_types[0]=='regression':
                _func1 = lambda x: x
            else:
                label2idx = {v:i for i,v in enumerate(encoder_classes[0])}
                _func1 = lambda x: label2idx[x]
                
            _func = partial(lambda_map_batch,
                           feature=self.label_names[0],
                           func=_func1,
                           output_feature='label',
                           is_batched=self.is_batched)
        return _func
        
    def _encode_labels(self):
        if len(self.label_names)==0: return
        print_msg('Label Encoding',verbose=self.verbose)
        
        if len(self.label_lists) and not isinstance(self.label_lists[0],list):
            self.label_lists = [self.label_lists]
                    
        encoder_classes=[]
        if not self.is_multilabel:
            for idx,l in enumerate(self.label_names):
                if self.sup_types[idx]=='regression':
                    l_classes=[]
                else:
                    # classification
                    if len(self.label_lists)==0:
                        l_encoder = LabelEncoder()
                        _ = l_encoder.fit(self.dset[l])
                        l_classes = list(l_encoder.classes_)
                    else:
                        l_classes = sorted(list(self.label_lists[idx]))
                encoder_classes.append(l_classes)
            
            _func = self._create_label_mapping_func(encoder_classes)
                
            self.dset = hf_map_dset(self.dset,_func,self.is_batched,self.batch_size,self.num_proc)
            if self.val_key is not None:
                self.ddict_rest[self.val_key] = hf_map_dset(self.ddict_rest[self.val_key],_func,
                                                       self.is_batched,self.batch_size,self.num_proc)
                    
        else:
            # For MultiLabel, we transform the label itself to one-hot (or actually, few-hot)
            if len(self.label_lists)==0:
                l_encoder = MultiLabelBinarizer()
                _ = l_encoder.fit(self.dset[self.label_names[0]])
                l_classes = list(l_encoder.classes_)
            else:
                l_classes = sorted(list(self.label_lists[0]))
            
            encoder_classes.append(l_classes)
            
            l_encoder = MultiLabelBinarizer(classes=encoder_classes[0])
            _ = l_encoder.fit(None)
            _func = partial(lambda_map_batch,
                            feature=self.label_names[0],
                            func=lambda x: l_encoder.transform(x),
                            output_feature='label',
                            is_batched=self.is_batched,
                            is_func_batched=True)
            self.dset = hf_map_dset(self.dset,_func,self.is_batched,self.batch_size,self.num_proc)                                                 
            
            if self.val_key is not None:
                self.ddict_rest[self.val_key] = hf_map_dset(self.ddict_rest[self.val_key],_func,
                                                       self.is_batched,self.batch_size,self.num_proc)
            
        self.label_lists = encoder_classes
        self.verboseprint('Done')
        
    
    def _train_test_split(self):
        print_msg('Train Test Split',20,verbose=self.verbose)
        if self.val_key is not None: # val split exists
            self.verboseprint('Validation split already exists')
            self.main_ddict=DatasetDict({'train':self.dset,
                                         'validation':self.ddict_rest.pop(self.val_key)})
            
    
        elif self.val_ratio is None: # use all data
            self.verboseprint('No validation split defined')
            self.main_ddict=DatasetDict({'train':self.dset})
            
        elif (isinstance(self.val_ratio,float) or isinstance(self.val_ratio,int)) and len(self.stratify_cols)==0:
            self.verboseprint('Validation split based on val_ratio')
            # train val split
            self.main_ddict = self.dset.train_test_split(test_size=self.val_ratio,shuffle=True,seed=self.seed)
            self.main_ddict['validation']=self.main_ddict['test']
            del self.main_ddict['test']
        
        else: # val_ratio split with stratifying             
            if self.is_multilabel and self.label_names[0] in self.stratify_cols:
                raise ValueError('For MultiLabel classification, you cannot choose the label as your stratified column')
            self.verboseprint('Validation split based on val_ratio, with stratifying')
            # Create a new feature 'stratified', which is a concatenation of values in stratify_cols
            if self.is_batched:
                stratified_creation = lambda x: {'stratified':
                                     ['_'.join(list(map(str,[x[v][i] for v in self.stratify_cols]))) 
                                      for i in range(len(x[self.stratify_cols[0]]))]}
            else:
                stratified_creation = lambda x: {'stratified':
                                     '_'.join(list(map(str,[x[v] for v in self.stratify_cols]))) 
                                      }
            self.dset = self.dset.map(stratified_creation,
                                      batched=self.is_batched,
                                      batch_size=self.batch_size,
                                      num_proc=self.num_proc)
            
            self.dset=self.dset.class_encode_column("stratified")
            # train val split
            self.main_ddict = self.dset.train_test_split(test_size=self.val_ratio,
                                                         shuffle=True,seed=self.seed,
                                                        stratify_by_column='stratified')
            self.main_ddict['validation']=self.main_ddict['test']
            del self.main_ddict['test']
            self.main_ddict=self.main_ddict.remove_columns(['stratified'])
            
        
        del self.dset
        self.verboseprint('Done')
    
    def _simplify_ddict(self):
        print_msg('Dropping unused features',20,verbose=self.verbose)
        if self.cols_to_keep is None:
            self.cols_to_keep= [self.main_text] + self.metadatas + self.label_names
            
        cols_to_remove = set(self.all_cols) - set(self.cols_to_keep)
        self.main_ddict['train']=self.main_ddict['train'].remove_columns(list(cols_to_remove))
        if 'validation' in self.main_ddict.keys():
            self.main_ddict['validation']=self.main_ddict['validation'].remove_columns(list(cols_to_remove))
        self.verboseprint('Done')
    
    def _do_transformation(self,dset,ddict_rest=None):
        if len(self.content_tfms):
            print_msg('Text Transformation',20,verbose=self.verbose)
            for tfm in self.content_tfms:
                print_msg(callable_name(tfm),verbose=self.verbose)
                _func = partial(lambda_map_batch,
                               feature=self.main_text,
                               func=tfm,
                               is_batched=self.is_batched)
                dset = hf_map_dset(dset,_func,self.is_batched,self.batch_size,self.num_proc)
                if ddict_rest is not None:
                    ddict_rest = hf_map_dset(ddict_rest,_func,self.is_batched,self.batch_size,self.num_proc)
            self.verboseprint('Done')
        return dset if ddict_rest is None else (dset,ddict_rest)
 
    def _do_filtering(self,dset,ddict_rest=None):
        if len(self.filter_dict):
            print_msg('Data Filtering',20,verbose=self.verbose)
            col_names = get_dset_col_names(dset)
            for f,tfm in self.filter_dict.items():
                if f in col_names:
                    print_msg(f'Do {callable_name(tfm)} on {f}',verbose=self.verbose)
                    _func = partial(lambda_batch,
                                    feature=f,
                                    func=tfm,
                                    is_batched=self.is_batched)
                    dset = hf_filter_dset(dset,_func,self.is_batched,self.batch_size,self.num_proc)
                if ddict_rest is not None: # assuming ddict_rest has the column to filter, always
                    ddict_rest = hf_filter_dset(ddict_rest,_func,self.is_batched,self.batch_size,self.num_proc)
            self.verboseprint('Done')
        return dset if ddict_rest is None else (dset,ddict_rest)
    
    def _upsampling(self):
        if len(self.upsampling_list):
            print_msg('Upsampling data',20,verbose=self.verbose)
            results=[]
            for f,tfm in self.upsampling_list:
                print_msg(f'Do {callable_name(tfm)} on {f}',verbose=self.verbose)
                _func = partial(lambda_batch,
                                feature=f,
                                func=tfm,
                                is_batched=self.is_batched)
                new_dset = hf_filter_dset(self.main_ddict['train'],_func,self.is_batched,self.batch_size,self.num_proc)
                results.append(new_dset)
            # slow concatenation for iterable dataset    
            self.main_ddict['train'] = concatenate_datasets(results+[self.main_ddict['train']])
            self.verboseprint('Done')
      
    def _do_augmentation(self):
        if len(self.aug_tfms):
            print_msg('Text Augmentation',20,verbose=self.verbose)
            for tfm in self.aug_tfms:
                print_msg(callable_name(tfm),verbose=self.verbose)
                bs = self.batch_size
                is_func_batched=False
                num_proc = self.num_proc
                is_batched = self.is_batched
                if hasattr(tfm, "run_on_gpu") and getattr(tfm,'run_on_gpu')==True:
                    bs = min(32,self.batch_size) if not hasattr(tfm, "batch_size") else getattr(tfm,'batch_size')
                    is_func_batched=True
                    is_batched=True
                    num_proc=1

                _func = partial(lambda_map_batch,
                               feature=self.main_text,
                               func=tfm,
                               is_batched=is_batched,
                               is_func_batched=is_func_batched
                               )
                self.main_ddict['train'] = hf_map_dset(self.main_ddict['train'],_func,
                                                          is_batched=is_batched,
                                                          batch_size=bs,
                                                          num_proc=num_proc
                                                         )


            self.verboseprint('Done')
        

            
    def _do_train_shuffling(self):
        print_msg('Shuffling and flattening train set',20,verbose=self.verbose)
        self.main_ddict['train'] = self.main_ddict['train'].shuffle(seed=self.seed).flatten_indices(num_proc = self.num_proc)
        self.verboseprint('Done')
        
    def do_all_preprocessing(self,
                             shuffle_trn=True # To shuffle the train set before tokenization
                            ):
        if self._processed_call:
            warnings.warn('Your dataset has already been processed. Returning the previous processed DatasetDict...')
            return self.main_ddict
            
        print_msg('Start Main Text Processing',20,verbose=self.verbose)
        
        # Filtering
        self.dset,self.ddict_rest = self._do_filtering(self.dset,self.ddict_rest)
        
        # Process metadatas
        self.dset,self.ddict_rest = self._process_metadatas(self.dset,self.ddict_rest)
        
        # Label transformation
        self._do_label_transformation()
        
        # Process labels
        self._encode_labels()
        
        # Content transformation
        self.dset,self.ddict_rest = self._do_transformation(self.dset,self.ddict_rest)
         
        # Train Test Split.
        ### self.main_ddict is created here
        self._train_test_split()
        
        # Dropping unused columns
        self._simplify_ddict()
        
        # Check validation leaking
        self._check_validation_leaking()
        
        ### The rest of these functions applies only to the train dataset
        # Upsampling
        self._upsampling()
        
        # Augmentation
        self._do_augmentation()
        
        # Shuffle train
        if shuffle_trn:
            self._do_train_shuffling()
        
        self._processed_call=True
        
        return self.main_ddict
    
        
    def do_tokenization(self,
                        tokenizer, # Tokenizer (preferably from HuggingFace)
                        max_length=None, # pad to model's allowed max length (default is max_sequence_length). Use -1 for no padding at all
                        trn_size=None, # The number of training data to be tokenized
                        tok_num_proc=None, # Number of processes for tokenization
                       ):
        print_msg('Tokenization',20,verbose=self.verbose)
        self.tokenizer = tokenizer
        self.max_length = max_length
        self.tok_num_proc = tok_num_proc if tok_num_proc else self.num_proc
        tok_func = partial(tokenize_function,tok=tokenizer,max_length=max_length)
        _func = partial(lambda_map_batch,
                        feature=self.main_text,
                        func=tok_func,
                        output_feature=None,
                        is_batched=self.is_batched)
        
        if trn_size is not None:
            if isinstance(trn_size,float):
                num_shard = int(1/trn_size)
            else: # int
                trn_len=len(self.main_ddict['train'])
                num_shard = trn_len//trn_size
            self.main_ddict['train'] = self.main_ddict['train'].shard(num_shard,0)
        
        for k in self.main_ddict.keys():
            self.main_ddict[k] = hf_map_dset(self.main_ddict[k],_func,self.is_batched,self.batch_size,self.tok_num_proc)

        self.verboseprint('Done')
        return self.main_ddict
        
    def process_and_tokenize(self,
                             tokenizer, # Tokenizer (preferably from HuggingFace)
                             max_length=None, # pad to model's allowed max length (default is max_sequence_length)
                             trn_size=None, # The number of training data to be tokenized
                             tok_num_proc=None, # Number of processes for tokenization
                             shuffle_trn=True, # To shuffle the train set before tokenization
                            ):
        """
        This will perform `do_all_processing` then `do_tokenization`
        """
        if self.seed:
            seed_everything(self.seed) 
        _ = self.do_all_preprocessing(shuffle_trn)
        _ = self.do_tokenization(tokenizer,max_length,trn_size,tok_num_proc)
        
    
    def set_data_collator(self,data_collator):
        self.data_collator = data_collator
        
    
    def prepare_test_dataset_from_csv(self,
                                      file_path, # path to csv file
                                      do_filtering=False # whether to perform data filtering on this test set
                                     ):
        file_path = Path(file_path)
        ds = load_dataset(str(file_path.parent),
                          data_files=file_path.name,
                          split='train')
        return self.prepare_test_dataset(ds,do_filtering)
    
    def prepare_test_dataset_from_df(self,
                                     df, # Pandas Dataframe
                                     validate=True, # whether to perform input data validation
                                     do_filtering=False # whether to perform data filtering on this test set 
                                    ):
        if validate:
            check_input_validation(df)
        ds = Dataset.from_pandas(df)
        return self.prepare_test_dataset(ds,do_filtering)
    
    def prepare_test_dataset_from_raws(self,
                                       content, # Either a single sentence, list of sentence or a dictionary with keys are metadata columns and values are list
                                      ):
        if len(self.metadatas) and not isinstance(content,dict):
            raise ValueError(f'There is/are metadatas in the preprocessing step. Please include a dictionary including these keys for metadatas: {self.metadatas}, and texture content: {self.main_text}')
            
        _dic = {self.main_text:[content]} if isinstance(content,str) else content
        for k in _dic.keys():
            _dic[k] = val2iterable(_dic[k])
        
        test_dict = Dataset.from_dict(_dic)
        
        # set num_proc to 1 for small data processing
        _tmp1 = self.num_proc
        _tmp2 = self.tok_num_proc
        self.num_proc=1
        self.tok_num_proc=1
        results = self.prepare_test_dataset(test_dict,do_filtering=False)
        self.num_proc = _tmp1
        self.tok_num_proc=_tmp2

        return results
    
    def prepare_test_dataset(self,
                             test_dset, # The HuggingFace Dataset as Test set
                             do_filtering=False, # whether to perform data filtering on this test set
                            ):
        test_cols = set(get_dset_col_names(test_dset))
        label_names_set = set(self.label_names)
        test_cols = test_cols - label_names_set
        missing_cols = set(self.cols_to_keep) - label_names_set - test_cols
        if len(missing_cols):
            raise ValueError(f'Test set does not have these columns required for preprocessings: {missing_cols}')
            
        print_msg('Start Test Set Transformation',20,verbose=self.verbose)

        # Filtering
        if do_filtering:
            test_dset = self._do_filtering(test_dset)
        
        # Process metadatas
        test_dset = self._process_metadatas(test_dset)
        
        # Content transformation
        test_dset = self._do_transformation(test_dset)
        
        # Drop unused columns
        cols_to_remove = test_cols - set(self.cols_to_keep)
        test_dset=test_dset.remove_columns(list(cols_to_remove))
        
        # Tokenization
        print_msg('Tokenization',20,verbose=self.verbose)
        _func = partial(lambda_map_batch,
                        feature=self.main_text,
                        func=partial(tokenize_function,tok=self.tokenizer,max_length=self.max_length),
                        output_feature=None,
                        is_batched=self.is_batched)
        test_dset = hf_map_dset(test_dset,_func,self.is_batched,self.batch_size,self.tok_num_proc)
        
        self.verboseprint('Done')
        return test_dset




# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/04_models.roberta.classifiers.ipynb.

# %% ../../../nbs/04_models.roberta.classifiers.ipynb 3
from __future__ import annotations
import torch
from transformers.models.roberta.configuration_roberta import RobertaConfig
from transformers.models.roberta.modeling_roberta import RobertaModel
from transformers.models.roberta.modeling_roberta import RobertaPreTrainedModel
from transformers.modeling_outputs import SequenceClassifierOutput
from ...model_main import loss_for_classification
from ...utils import *

# %% auto 0
__all__ = ['ConcatHeadExtended', 'ConcatHeadSimple', 'RobertaClassificationHeadCustom', 'RobertaBaseForSequenceClassification',
           'RobertaHiddenStateConcatForSequenceClassification']

# %% ../../../nbs/04_models.roberta.classifiers.ipynb 5
class ConcatHeadExtended(torch.nn.Module):
    """
    Concatenated head for Roberta Classification Model. 
    This head takes the last n hidden states of [CLS], and concatenate them before passing through the classifier head
    """
    def __init__(self,
                 config, # HuggingFace model configuration
                 classifier_dropout=0.1, # Dropout ratio (for dropout layer right before the last nn.Linear)
                 last_hidden_size=768, # Last hidden size (before the last nn.Linear)
                 layer2concat=4, # number of hidden layer to concatenate (counting from top)
                 num_labels=None, # Number of label output. Overwrite config.num_labels 
                 **kwargs
                ):

        super().__init__()
        self.last_hidden_size=last_hidden_size
        self.dropout = torch.nn.Dropout(classifier_dropout)
        self.pre_classifier = torch.nn.Linear(layer2concat*config.hidden_size,last_hidden_size)
        num_labels=num_labels if num_labels is not None else config.num_labels
        self.out_proj = torch.nn.Linear(last_hidden_size, num_labels)
    
    def forward(self, inp, **kwargs):
        x = inp
        x = self.dropout(x)
        x = self.pre_classifier(x)
        x = torch.tanh(x)
#         x = torch.relu(x)
        x = self.dropout(x)
        x = self.out_proj(x)
        return x

# %% ../../../nbs/04_models.roberta.classifiers.ipynb 7
class ConcatHeadSimple(torch.nn.Module):
    """
    Concatenated head for Roberta Classification Model, the simpler version (no hidden linear layer)
    This head takes the last n hidden states of [CLS], and concatenate them before passing through the classifier head
    """
    def __init__(self,
                 config, # HuggingFace model configuration
                 classifier_dropout=0.1, # Dropout ratio (for dropout layer right before the last nn.Linear)
                 layer2concat=4, # number of hidden layer to concatenate (counting from top)
                 num_labels=None, # Number of label output. Overwrite config.num_labels 
                 **kwargs
                ):

        super().__init__()
        self.dropout = torch.nn.Dropout(classifier_dropout)
        num_labels=num_labels if num_labels is not None else config.num_labels
        self.out_proj = torch.nn.Linear(layer2concat*config.hidden_size, num_labels)
    def forward(self, inp, **kwargs):
        x = inp
        x = self.dropout(x)
        x = self.out_proj(x)
        return x

# %% ../../../nbs/04_models.roberta.classifiers.ipynb 9
class RobertaClassificationHeadCustom(torch.nn.Module):
    """
    Same as RobertaClassificationHead, but you can freely adjust dropout
    
    Reference: https://github.com/huggingface/transformers/blob/main/src/transformers/models/roberta/modeling_roberta.py#L1424
    """
    
    def __init__(self, 
                 config, # HuggingFace model configuration
                 classifier_dropout=0.1, # Dropout ratio (for dropout layer right before the last nn.Linear)
                 num_labels=None, # Number of label output. Overwrite config.num_labels 
                 **kwargs
                ):
        super().__init__()
        self.dense = torch.nn.Linear(config.hidden_size, config.hidden_size)
        self.dropout = torch.nn.Dropout(classifier_dropout)
        num_labels=num_labels if num_labels is not None else config.num_labels
        self.out_proj = torch.nn.Linear(config.hidden_size, num_labels)

    def forward(self, features, **kwargs):
        x = self.dropout(x)
        x = self.dense(x)
        x = torch.tanh(x)
        x = self.dropout(x)
        x = self.out_proj(x)
        return x

# %% ../../../nbs/04_models.roberta.classifiers.ipynb 12
class RobertaBaseForSequenceClassification(RobertaPreTrainedModel):
    """
    Base Roberta Architecture for Sequence Classification task
    
    Based on: https://github.com/huggingface/transformers/blob/main/src/transformers/models/roberta/modeling_roberta.py#L1155C35-L1155C35
    """
    # make sure standard XLM-R are used
    config_class = RobertaConfig

    def __init__(self,
                 config, # HuggingFace model configuration
                 is_multilabel=False, # Whether this is a multilabel classification
                 is_multihead=False, # Whether this is a multihead (multi-level) classification
                 head_class_sizes=[], # Class size for each head
                 head_weights=[], # loss weight for each head. This will be multiplied to the loss of each head's output
                 head_class=None, # The class object of the head. You can use RobertaClassificationHeadCustom as default
                 **head_class_kwargs, # Keyword arguments for the head class
                ):
        super().__init__(config)
        self.is_multilabel = is_multilabel
        self.is_multihead = is_multihead
        self.head_class_sizes = val2iterable(head_class_sizes)
        self.head_weights = val2iterable(head_weights,lsize=len(self.head_class_sizes))
        
        # set num_labels for config
        num_labels = sum(self.head_class_sizes)
        config.num_labels = num_labels
        
        # add_pooling_layer to False to ensure all hidden states are returned and not only the one associated with the [CLS] token.
        self.body_model = RobertaModel(config, add_pooling_layer=False)
        # Set up head
        self.head = head_class(config=config,**head_class_kwargs)


    def forward(
        self,
        input_ids= None,
        attention_mask= None,
        token_type_ids= None,
        position_ids= None,
        head_mask= None,
        inputs_embeds= None,
        labels= None,
        output_attentions= None,
        output_hidden_states= None,
        return_dict= None,
        ):
        
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict
        output_hidden_states = output_hidden_states if output_hidden_states is not None else self.config.output_hidden_states

        outputs = self.body_model(
            input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
            head_mask=head_mask,
            inputs_embeds=inputs_embeds,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )
        sequence_output = outputs[0]
        sequence_output = sequence_output[:, 0, :]  # take <s> token (equiv. to [CLS])
        logits = self.head(sequence_output) # (bs,sum of all class sizes)
        
        # Calculate losses
        if labels is None:
            loss=None
        else:
            loss = loss_for_classification(logits, labels, 
                                   self.is_multilabel,
                                   self.is_multihead, 
                                   self.head_class_sizes,
                                   self.head_weights)
        
        if not return_dict:
            output = (logits,) + outputs[2:]
            return ((loss,) + output) if loss is not None else output

        return SequenceClassifierOutput(loss=loss, logits=logits,
#                                      hidden_states=outputs.hidden_states,
                                        hidden_states=None,
                                        attentions=outputs.attentions)

# %% ../../../nbs/04_models.roberta.classifiers.ipynb 14
class RobertaHiddenStateConcatForSequenceClassification(RobertaPreTrainedModel):
    """
    Roberta Architecture with Hidden-State-Concatenation for Sequence Classification task
    """
    
    config_class = RobertaConfig

    def __init__(self,config, # HuggingFace model configuration
                 layer2concat=4, # number of hidden layer to concatenate (counting from top)
                 is_multilabel=False, # Whether this is a multilabel classification
                 is_multihead=False, # Whether this is a multihead (multi-level) classification
                 head_class_sizes=[], # Class size for each head
                 head_weights=[], # loss weight for each head. This will be multiplied to the loss of each head's output
                 head_class=None, # The class object of the head. You can use RobertaClassificationHeadCustom as default
                 **head_class_kwargs, # Keyword arguments for the head class
                ):
        super().__init__(config)
        self.is_multilabel = is_multilabel
        self.is_multihead = is_multihead
        self.head_class_sizes = val2iterable(head_class_sizes)
        self.head_weights = val2iterable(head_weights,lsize=len(self.head_class_sizes))
        self.layer2concat=layer2concat
        # set num_labels for config
        num_labels = sum(self.head_class_sizes)
        config.num_labels = num_labels
        
        # Load model body
        # add_pooling_layer to False to ensure all hidden states are returned  and not only the one associated with the [CLS] token.
        self.body_model = RobertaModel(config, add_pooling_layer=False)
        
        # Set up head
        self.head = head_class(config=config,layer2concat=layer2concat,
                                              **head_class_kwargs)

    def forward(
        self,
        input_ids= None,
        attention_mask= None,
        token_type_ids= None,
        position_ids= None,
        head_mask= None,
        inputs_embeds= None,
        labels= None,
        output_attentions= None,
        output_hidden_states= None,
        return_dict= None,
        ):
        
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict
        output_hidden_states = output_hidden_states if output_hidden_states is not None else self.config.output_hidden_states

        outputs = self.body_model(
            input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
            head_mask=head_mask,
            inputs_embeds=inputs_embeds,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )
        
        hidden_states = outputs['hidden_states'] # tuples with len 13 (number of layer/block)
        # each with shape: (bs,seq_len,hidden_size_len), e.g. for phobert: (bs,256, 768)
        # Note: hidden_size_len = embedding_size
        hidden_concat = torch.cat([hidden_states[i][:,0] for i in range(-1,-self.layer2concat-1,-1)],
                                  -1) 
        logits = self.head(hidden_concat) # (bs,sum of all class sizes)
        
        # Calculate losses
        if labels is None:
            loss=None
        else:            
            loss = loss_for_classification(logits, labels, 
                                   self.is_multilabel,
                                   self.is_multihead, 
                                   self.head_class_sizes,
                                   self.head_weights)
            
        if not return_dict:
            output = (logits,) + outputs[2:]
            return ((loss,) + output) if loss is not None else output

        return SequenceClassifierOutput(loss=loss, logits=logits,
#                                      hidden_states=outputs.hidden_states,
                                        hidden_states=None,
                                        attentions=outputs.attentions)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/07_utils.ipynb.

# %% ../nbs/07_utils.ipynb 3
from __future__ import annotations
import os, sys
from pathlib import Path
from functools import partial, reduce
import dill as pickle
import torch
import numpy as np
import pandas as pd
from typing import Callable, Any
from collections.abc import Iterable
from datasets import IterableDataset
import random
import warnings

# %% auto 0
__all__ = ['HiddenPrints', 'val2iterable', 'create_dir', 'check_and_get_attribute', 'callable_name', 'print_msg', 'seed_notorch',
           'seed_everything', 'save_to_pickle', 'load_pickle', 'check_input_validation', 'check_text_leaking',
           'none2emptystr', 'lambda_batch', 'lambda_map_batch', 'augmentation_stream_generator', 'func_all',
           'get_dset_col_names', 'hf_map_dset', 'hf_filter_dset', 'resize_model_embeddings', 'sigmoid']

# %% ../nbs/07_utils.ipynb 4
class HiddenPrints:
    "To hide print command when called"
    def __enter__(self):
        self._original_stdout = sys.stdout
        sys.stdout = open(os.devnull, 'w')

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout.close()
        sys.stdout = self._original_stdout

# %% ../nbs/07_utils.ipynb 5
def val2iterable(val,lsize=1,t='list'):
    "Convert an element (nonlist value) to an iterable. Currently support list and nparray"
    if not isinstance(val, Iterable) or isinstance(val,str):
        if t=='list': val=[val for i in range(lsize)]
        elif t=='nparray': val=np.repeat(val,lsize)
        else:
            raise ValueError('Unrecognized iterable to convert to')
    return val

def create_dir(path_dir):
    "Create directory if needed"
    path_dir = Path(path_dir)
    if not path_dir.exists():
        path_dir.mkdir(parents=True)

# %% ../nbs/07_utils.ipynb 6
def check_and_get_attribute(obj,attr_name):
    if hasattr(obj,attr_name): return getattr(obj,attr_name)
    else: raise ValueError(f"Missing required argument: {attr_name}")
            
def callable_name(any_callable: Callable[..., Any]) -> str:
    "To get name of any callable"
    if hasattr(any_callable, '__name__'):
        return any_callable.__name__
    if isinstance(any_callable, partial):
        return any_callable.func.__name__

    return str(any_callable)

def print_msg(msg,dash_num=5,verbose=True):
    if verbose:
        print('-'*dash_num+' '+msg+' '+ '-'*dash_num)

def seed_notorch(seed=42):
    random.seed(seed)
    np.random.seed(seed)
    
def seed_everything(seed=42):
    seed_notorch(seed=seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.backends.cudnn.deterministic = True

# %% ../nbs/07_utils.ipynb 7
def save_to_pickle(my_list,fname,parent='pickle_files'):
    if fname[-4:]!='.pkl': fname+='.pkl'
    p = Path(parent)
    p.mkdir(parents=True, exist_ok=True)
    fname = p/fname
    with open(fname, 'wb') as f:
        pickle.dump(my_list, f)
        
def load_pickle(fname,parent='pickle_files'):
    if fname[-4:]!='.pkl': fname+='.pkl'
    with open(Path(parent)/fname,'rb') as f:
        my_list = pickle.load(f)
    return my_list

# %% ../nbs/07_utils.ipynb 8
def check_input_validation(df:pd.DataFrame,verbose=True):
    verboseprint = print if verbose else lambda *a, **k: None
    print_msg('Input Validation Precheck',verbose)
    # check whether index is sorted
    correct_idxs = np.arange(df.shape[0])
    curr_idxs = df.index.values
    if not np.array_equal(correct_idxs,curr_idxs):
        verboseprint("DataFrame Index is not RangeIndex, and will be converted")
        df.index = correct_idxs

    # Do a NA check:
    na_check = df.isna().sum(axis=0)
    na_check = na_check[na_check!=0]
    if na_check.shape[0]!=0:
        verboseprint("Data contains missing values!")
        verboseprint('-----> List of columns and the number of missing values for each')
        verboseprint(na_check)

    # Do a row duplication check
    _df = df.copy().astype(str)
    dup_check = _df.value_counts(dropna=False)
    dup_check=dup_check[dup_check>1]
    if dup_check.shape[0]!=0:
        verboseprint("Data contains duplicated values!")
        verboseprint(f'-----> Number of duplications: {(dup_check.values-1).sum()} rows')
        

# %% ../nbs/07_utils.ipynb 9
def check_text_leaking(trn_txt:list,
                       test_txt:list,verbose=True):
    verboseprint = print if verbose else lambda *a, **k: None
    test_txt_leaked = {i.strip().lower() for i in trn_txt} & {j.strip().lower() for j in test_txt}
    len_leaked = len(test_txt_leaked)
    verboseprint(f'- Number of rows leaked: {len_leaked}, which is {100*len_leaked/len(trn_txt):.2f}% of training set')
    return test_txt_leaked

# %% ../nbs/07_utils.ipynb 10
def none2emptystr(x):
    if x is None: return ''
    return str(x)

# %% ../nbs/07_utils.ipynb 11
def lambda_batch(inp, # HuggingFace Dataset
                 feature, # Feature name.
                 func, # The function to apply
                 is_batched, # Whether batching is applied
                ):
    return [func(v) for v in inp[feature]] if is_batched else func(inp[feature])

# %% ../nbs/07_utils.ipynb 12
def lambda_map_batch(inp, # HuggingFace Dataset
                     feature, # Feature name.
                     func, # The function to apply
                     is_batched, # Whether batching is applied
                     output_feature='same', # New feature output, if different from 'feature'. If none, use function's output feature(s)
                     is_func_batched=False # Whether the func above only works with batch (mostly sklearn's)
                    ):
    
    if output_feature is None: 
        return func(inp[feature]) if is_batched else func([inp[feature]])
    if output_feature.strip().lower()=='same': output_feature = feature
    results={}
    if not is_func_batched:
        results[output_feature] = lambda_batch(inp,feature,func,is_batched)
    else:
        results[output_feature] = func(inp[feature]) if is_batched else func([inp[feature]])
    return results

# %% ../nbs/07_utils.ipynb 13
def augmentation_stream_generator(dset,text_name,func):
    for inp in dset:
        # inp[text_name] will be a single item
        inp[text_name]=func(inp[text_name])
        yield inp

# %% ../nbs/07_utils.ipynb 15
def func_all(x, functions):
    return reduce(lambda acc, func: func(acc), functions, x)

# %% ../nbs/07_utils.ipynb 16
def get_dset_col_names(dset):
    if dset.column_names is not None: return dset.column_names
    warnings.warn("Iterable Dataset might contain multiple mapping functions; getting column names can be time and memory consuming") 
    return list(next(iter(dset)).keys())

# %% ../nbs/07_utils.ipynb 17
def hf_map_dset(dset,func,
                is_batched=True,
                batch_size=1024,
                num_proc=1):
    is_streamed = isinstance(dset,IterableDataset)
    if is_streamed:
        return dset.map(func,
                        batched=is_batched,
                        batch_size=batch_size
                       )
    return dset.map(func,
                    batched=is_batched,
                    batch_size=batch_size,
                    num_proc=num_proc
                   )

def hf_filter_dset(dset,func,
                   is_batched=True,
                   batch_size=1024,
                   num_proc=1):
    is_streamed = isinstance(dset,IterableDataset)
    if is_streamed:
        return dset.filter(func,
                           batched=is_batched,
                           batch_size=batch_size
                          )
    return dset.filter(func,
                       batched=is_batched,
                       batch_size=batch_size,
                       num_proc=num_proc
                      )

# %% ../nbs/07_utils.ipynb 18
def resize_model_embeddings(model,tokenizer):
    embedding_size = model.get_input_embeddings().weight.shape[0]
    if len(tokenizer) > embedding_size:
        model.resize_token_embeddings(len(tokenizer))
    return model

# %% ../nbs/07_utils.ipynb 19
def sigmoid(x):
    """
    A numerically stable version of the logistic sigmoid function.
    
    Source: assignment3 of cs231n
    """
    pos_mask = (x >= 0)
    neg_mask = (x < 0)
    z = np.zeros_like(x)
    z[pos_mask] = np.exp(-x[pos_mask])
    z[neg_mask] = np.exp(x[neg_mask])
    top = np.ones_like(x)
    top[neg_mask] = z[neg_mask]
    return top / (1 + z)
